# -*- coding: utf-8 -*-
"""
A dictionary like class that captures all the global settings used for an atmospheric evolution. 

@author: jeremy leconte
"""


class Settings(object):
    def __init__(self):
        """Initializes all global parameters to default values

        *Note for developers*: The initial `self.parameters` dictionary (the one below)
        should not contain any item whose key is a keyword argument for Atm.__init__().
        It should also contain every other possible option that does not require
        re-initializing the radiative transfer model. This is necessary for
        Atm_evolution.set_options() to know when to reset the radiative model. 
        """
        self.parameters={'rayleigh': True,
                        'convection': False,
                        'convective_transport': True,
                        'diffusion': False,
                        'molecular_diffusion': False,
                        'condensation': False,
                        'rain': False,
                        'latent_heating': True,
                        'moist_convection': False,
                        'moist_inhibition': False,
                        'surface_reservoir': False,
                        'mass_redistribution': False,
                        'compute_mass_fluxes': True,
                        'dTmax_use_kernel': 10.,
                        'evap_coeff': 1.,
                        'acceleration_mode': 0,
                        'radiative_acceleration_reducer': 1.,
                        'condensation_timestep_reducer': .8,
                        'convective_acceleration_mode': 0,
                        'qcond_surf_layer': 0.1,
                        'q_cloud': 0.,
                        }
            
        self._forbidden_changes = ['logplay', 'play']

        self._non_radiative_parameters = set(self.parameters.keys())

    def set_parameters(self, **kwargs):
        """Sets various global options
        """
        for key, val in kwargs.items():
            self.parameters[key]=val
        if 'logplay' in self.keys():
            self['Nlay']=self['logplay'].size

    def use_or_set(self, key, value):
        """Returns the value stored in the parameters if available.
        Stores the given value if not. 
        """
        if key in self.parameters.keys():
            return self.parameters[key]
        else:
            self.parameters[key] = value
            return value

    def __getitem__(self, param):
        return self.parameters[param]

    def __setitem__(self, key, item):
        self.parameters[key] = item

    def pop(self, key, default):
        return self.parameters.pop(key, default)

    def get(self, key, default):
        return self.parameters.get(key, default)

    def keys(self):
        return self.parameters.keys()

    def items(self):
        return self.parameters.items()

    def values(self):
        return self.parameters.values()

    def __repr__(self):
        """Method to output parameters
        """
        return self.parameters.__repr__()



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
import numba
import exo_k.util.cst as cst


class Condensing_species(object):

    def __init__(self, Latent_heat_vaporization = 0., cp_vap = 0., Mvap = 0.,
            T_ref = 0., Psat_ref = 0., cp_cond = None):
        """

        Parameters
        ----------
            Latent_heat_vaporization: float
                specific Latent heat vaporization (J/kg)
            cp_vap: float
                Specific heat capacity of the vapor (J/kg/K)
            Mvap: float
                Molar mass of vapor (kg/mol)
            T_ref: float
                Reference temperature
            Psat_ref: float
                Saturation vapor pressure at the reference temperature (Pa)
            cp_cond: float (optional)
                Specific heat capacity of the condensate (J/kg/K).
                Assumed equal to cp_vap if not provided.

        """
        self.Latent_heat_vaporization = Latent_heat_vaporization
        self.cp_vap = cp_vap
        self.T_ref = T_ref
        self.Psat_ref = Psat_ref
        self.Mvap = Mvap
        self.Rvap =  cst.RGP / self.Mvap
        if cp_cond is None:
            self.cp_cond = cp_vap
        else:
            self.cp_cond = cp_cond
        self.delta_cp = self.cp_vap - self.cp_cond 
        self.delta_cp_R = self.delta_cp / self.Rvap
        self.LovR = self.Latent_heat_vaporization / self.Rvap
        self.c1 = self.LovR / self.T_ref - self.delta_cp_R
        self.c2 = self.delta_cp_R * self.T_ref - self.LovR 
 
    def Lvap(self, T):
        """Latent heat at temperature T

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
        """
        return Lvap_T(T, self.Latent_heat_vaporization, self.T_ref, self.delta_cp)

    def Psat(self, T):
        """Saturation vapor pressure for the condensing species

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
        """
        return Psat_T(T, self.T_ref, self.Psat_ref, self.c1, self.c2, self.delta_cp_R)

    def Tsat(self, P):
        """Boiling temperature for the condensing species

        NOT EXACT IF DELTA CP != 0.

        Parameters
        ----------
            P: array, np.ndarray
                Pressure in layers (Pa)
        """
        return Tsat_P(P, self.Psat_ref, self.c1, self.c2)


    def qsat(self, psat, p, epsilon):
        """Saturation vapor mass mixing ratio for the condensing species

        Parameters
        ----------
            psat: array, np.ndarray
                saturation vapor pressure
            p: array, np.ndarray
                pressure at the layer center
            epsilon: float or array
                Ratio of the molar mass of the vapor over the background molar mass
        """
        return Qsat(psat, p, epsilon)

    def dPsat_dT(self, T):
        """Saturation vapor pressure derivative for the condensing species 

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
        """
        return dPsat_dT(T, self.Latent_heat_vaporization, self.T_ref,
            self.Psat_ref, self.Rvap, self.delta_cp, self.delta_cp_R, self.c1, self.c2)


    def dlnPsat_dlnT(self, T):
        """Saturation vapor pressure for the condensing species and its derivative 

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
        """
        return dlnPsat_dlnT(T, self.Latent_heat_vaporization, self.T_ref,
            self.delta_cp, self.Rvap)


    def moist_adiabat(self, T, P, cp, Mgas):
        """Computes the threshold thermal gradient (d ln T / d ln P) for a moist atmosphere.

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
            P: array, np.ndarray
                pressure at the layer center
            cp: float
                specific heat capacity at constant pressure of the background gas
            Mgas: float or array
                Molar mass of the background atmosphere
        
        Returns
        -------
            array
                Moist adiabat lapse rate
            Lvap: array, np.ndarray
                Latent heat at temperature T
            psat: array, np.ndarray
                Saturation vapor pressure for the condensing species (Pa)
            qsat: array, np.ndarray
                Saturation vapor mass mixing ratio for the condensing species
            dqsat_dt: array, np.ndarray
                Derivative of qsat with respect to temperature at fixed pressure
            q_crit: array, np.ndarray
                Critical mass mixing ratio for the inhibition of moist convection
                (Eq. 17 of Leconte et al. 2017)

        """
        return moist_adiabat(T, P, cp, Mgas, self.cp_vap, self.Mvap, self.Rvap,
            self.Latent_heat_vaporization, self.T_ref, self.Psat_ref,
            self.delta_cp, self.delta_cp_R, self.c1, self.c2)


    def compute_condensation_parameters(self, T, P, Mgas):
        """Computes necessary quantities to compute
        large scale condensation.

        Parameters
        ----------
            T: array, np.ndarray
                Temperature in layers (K)
            P: array, np.ndarray
                pressure at the layer center
            Mgas: float or array
                Molar mass of the background atmosphere
        """
        return compute_condensation_parameters(T, P, Mgas, self.Mvap,
            self.Rvap, self.Latent_heat_vaporization, self.T_ref,
            self.Psat_ref, self.delta_cp, self.delta_cp_R, self.c1, self.c2)



    def __repr__(self):
        """Method to output parameters
        """
        output="""
        Lvap         : {lvap}
        cp (vap)     : {cp}
        T0        (K): {t0}
        Psat(T0) (Pa): {p0}""".format(lvap=self.Latent_heat_vaporization,
            cp=self.cp_vap, t0=self.T_ref, p0=self.Psat_ref)
        return output

class Condensation_Thermodynamical_Parameters(object):

    def __init__(self, Latent_heat_vaporization = 0., cp_vap = 0., Mvap = 0.,
            T_ref = 0., Psat_ref = 0., cp_cond = None):
        """

        Parameters
        ----------
            Latent_heat_vaporization: float
                specific Latent heat vaporization (J/kg)
            cp_vap: float
                Specific heat capacity of the vapor (J/kg/K)
            Mvap: float
                Molar mass of vapor (kg/mol)
            T_ref: float
                Reference temperature
            Psat_ref: float
                Saturation vapor pressure at the reference temperature (Pa)
            cp_cond: float (optional)
                Specific heat capacity of the condensate (J/kg/K).
                Assumed equal to cp_vap if not provided.

        """
        Rvap =  cst.RGP / Mvap
        if cp_cond is None:
            cp_cond = cp_vap
        else:
            cp_cond = cp_cond
        delta_cp = cp_vap - cp_cond 
        delta_cp_R = delta_cp / Rvap
        LovR = Latent_heat_vaporization / Rvap
        c1 = LovR / T_ref - delta_cp_R
        c2 = delta_cp_R * T_ref - LovR 
        self.th_params = np.array([cp_vap, Mvap, Rvap,
                Latent_heat_vaporization, T_ref, Psat_ref,
                delta_cp, delta_cp_R, c1, c2])

@numba.jit(nopython=True, fastmath=True, cache=True)
def Lvap_T(T, Latent_heat_vaporization, T_ref, delta_cp):
    """Latent heat at temperature T

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
    """
    return Latent_heat_vaporization + delta_cp * (T-T_ref)

@numba.jit(nopython=True, fastmath=True, cache=True)
def Psat_T(T, T_ref, Psat_ref, c1, c2, delta_cp_R):
    """Saturation vapor pressure for the condensing species

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
    """
    return Psat_ref*np.exp( c1 + c2/T + delta_cp_R*np.log(T/T_ref))

@numba.jit(nopython=True, fastmath=True, cache=True)
def Tsat_P(P, Psat_ref, c1, c2):
    """Boiling temperature for the condensing species

    NOT EXACT IF DELTA CP != 0.
    
    Parameters
    ----------
        P: array, np.ndarray
            Pressure in layers (Pa)
    """
    return c2/(np.log(P/Psat_ref)-c1)

@numba.jit(nopython=True, fastmath=True, cache=True)
def Qsat(psat, p, epsilon):
    """Saturation vapor mass mixing ratio for the condensing species

    Parameters
    ----------
        psat: array, np.ndarray
            saturation vapor pressure
        p: array, np.ndarray
            pressure at the layer center
        epsilon: float or array
            Ratio of the molar mass of the vapor over the background molar mass
    """
    psat_tmp = np.core.umath.minimum(psat,p)
    fac =  p + (epsilon -1.) * psat_tmp
    qsat = epsilon * psat_tmp / fac
    return qsat

@numba.jit(nopython=True, fastmath=True, cache=True)
def dPsat_dT(T, Latent_heat_vaporization, T_ref, Psat_ref, Rvap, delta_cp, delta_cp_R, c1, c2):
    """Saturation vapor pressure derivative for the condensing species 

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
    """
    RT2 = Rvap * T * T
    Lvap = Lvap_T(T, Latent_heat_vaporization, T_ref, delta_cp)
    psat = Psat_T(T, T_ref, Psat_ref, c1, c2, delta_cp_R)
    return psat * Lvap / RT2, psat, Lvap

@numba.jit(nopython=True, fastmath=True, cache=True)
def dlnPsat_dlnT(T, Latent_heat_vaporization, T_ref, delta_cp, Rvap):
    """Saturation vapor pressure for the condensing species and its derivative

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
    """
    RT = Rvap * T
    Lvap = Lvap_T(T, Latent_heat_vaporization, T_ref, delta_cp)
    return Lvap / RT, Lvap

@numba.jit(nopython=True, fastmath=True, cache=True)
def moist_adiabat(T, P, cp, Mgas, cp_vap, Mvap, Rvap, Latent_heat_vaporization, T_ref, Psat_ref, delta_cp, delta_cp_R, c1, c2):
    """Computes the threshold thermal gradient (d ln T / d ln P) for a moist atmosphere.

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
        P: array, np.ndarray
            pressure at the layer center
        cp: float
            specific heat capacity at constant pressure of the background gas
        Mgas: float or array
            Molar mass of the background atmosphere
    
    Returns
    -------
        array
            Moist adiabat lapse rate
        Lvap: array, np.ndarray
            Latent heat at temperature T
        psat: array, np.ndarray
            Saturation vapor pressure for the condensing species (Pa)
        qsat: array, np.ndarray
            Saturation vapor mass mixing ratio for the condensing species
        dqsat_dt: array, np.ndarray
            Derivative of qsat with respect to temperature at fixed pressure
        q_crit: array, np.ndarray
            Critical mass mixing ratio for the inhibition of moist convection
            (Eq. 17 of Leconte et al. 2017)
    """
    epsilon = Mvap / Mgas
    psat = Psat_T(T, T_ref, Psat_ref, c1, c2, delta_cp_R)
    qsat = Qsat(psat, P, epsilon)
    dlpsat_dlT, Lvap = dlnPsat_dlnT(T, Latent_heat_vaporization, T_ref, delta_cp, Rvap)
    dqsat_dt = qsat**2 * P * dlpsat_dlT / (epsilon*psat*T)
    qa=1.-qsat
    qLvt = qsat * Lvap / T 
    fac = qsat * cp_vap + qa * cp + qLvt * dlpsat_dlT
    q_crit = epsilon / ((epsilon - 1.) * dlpsat_dlT)
    return ( (1.-qsat) * cst.RGP / Mgas + qLvt ) / fac, Lvap, psat, qsat, dqsat_dt, q_crit # is missing p/p_a terms

@numba.jit(nopython=True, fastmath=True, cache=True)
def compute_condensation_parameters(T, P, Mgas, Mvap, Rvap, Latent_heat_vaporization, T_ref, Psat_ref, delta_cp, delta_cp_R, c1, c2):
    """Computes necessary quantities to compute
    large scale condensation.

    Parameters
    ----------
        T: array, np.ndarray
            Temperature in layers (K)
        P: array, np.ndarray
            pressure at the layer center
        Mgas: float or array
            Molar mass of the background atmosphere
    """
    epsilon = Mvap / Mgas
    psat = Psat_T(T, T_ref, Psat_ref, c1, c2, delta_cp_R)
    qsat = Qsat(psat, P, epsilon)
    dlpsat_dlT, Lvap = dlnPsat_dlnT(T, Latent_heat_vaporization, T_ref, delta_cp, Rvap)
    dqsat_dt = qsat**2 * P * dlpsat_dlT / (epsilon*psat*T)
    return Lvap, qsat, dqsat_dt


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

This module contain classes to handle the radiative properties of atmospheres.
This alows us to compute the transmission and emission spectra
of those atmospheres.

Physical properties of the atmosphere are handled in atm_profile.py.

The nomenclature for layers, levels, etc, is as follows (example with 4 layers; Nlay=4):

::

    -------------------  Model top or first level (plev[0])
    - - - - - - - - - -  First atmopheric layer (play[0]=plev[0], tlay[0], xlay[0])

    -------------------  plev[1]

    - - - - - - - - - -  play[1], tlay[1], xlay[1]

    -------------------  plev[2]

    - - - - - - - - - - play[2], tlay[2], xlay[2]

    -------------------  plev[3]

    - - - - - - - - - -  bottom layer (play[Nlay-1]=psurf, tlay[3], xlay[3])
    ------------------- Surface (plev[Nlev-1=Nlay]=psurf)
    ///////////////////

.. image:: /images/atm_schema.png


Temperatures (`tlay`) and volume mixing ratios (`xlay`) are provided at the
*mid layer* point (`play`) (Note that this does not mean
that this point is the middle of the layer, in general, it is not).

If pressure levels are not specified by the user with `logplevel`,
they are at the mid point between
the pressure of the layers directly above and below. The pressure of the
top and bottom levels are counfounded with the pressure of the
top and bottom mid layer points.

For radiative calculations, the source function (temperature) needs to be known at
the boudaries of the *radiative* layers but the opacity needs to be known
inside the *radiative* layer.
For this reason, there are `Nlay-1` radiative layers and they are offset
with respect to atmospheric layers.
Opacities are computed at the center of those radiative layers, i.e. at the pressure levels.
The temperature is interpolated at these levels with an arithmetic average.
Volume mixing ratios are interpolated using a geometric average. 

"""
from typing import Optional

import numpy as np
import warnings
from .atm_profile import Atm_profile
from .util.cst import PI, SIG_SB
from .util.radiation import Bnu_integral_num, Bnu, rad_prop_corrk, rad_prop_xsec,\
    Bnu_integral_array, path_integral_corrk, path_integral_xsec
from .two_stream import two_stream_toon as toon # noqa
from .two_stream import two_stream_lmdz as lmdz # noqa
from .util.interp import gauss_legendre
from .util.spectrum import Spectrum


class Atm(Atm_profile):
    """Class based on Atm_profile that handles radiative trasnfer calculations.

    Radiative data are accessed through the :any:`gas_mix.Gas_mix` class.
    """

    def __init__(self, k_database = None, cia_database = None, a_database = None,
        wn_range = None, wl_range = None, internal_flux = 0., rayleigh = False,
        flux_top_dw = 0., Tstar = 5570., albedo_surf = 0., wn_albedo_cutoff = 5000.,
        **kwargs):
        """Initialization method that calls Atm_Profile().__init__(**kwargs) and links
        to Kdatabase and other radiative data. 
        """
        super().__init__(**kwargs)
        self.known_spectral_axis = False
        self.number_computation_albedo_surf_nu = 0
        self.number_computation_flux_top_dw_nu = 0
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)
        self.set_a_database(a_database)
        self.set_spectral_range(wn_range = wn_range, wl_range = wl_range)
        self.set_internal_flux(internal_flux)

        self.compute_flux_top_dw_nu = True
        self.set_incoming_stellar_flux(flux_top_dw = flux_top_dw, Tstar = Tstar, **kwargs)

        self.compute_albedo_surf_nu = True
        self.set_surface_albedo(albedo_surf = albedo_surf,
                        wn_albedo_cutoff = wn_albedo_cutoff)
        self.set_rayleigh(rayleigh = rayleigh)
        self.flux_net_nu=None
        self.kernel=None

    def set_k_database(self, k_database=None):
        """Change the radiative database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_k_database` for details.

        Parameters
        ----------
            k_database: :class:`Kdatabase` object
                New Kdatabase to use.
        """
        self.gas_mix.set_k_database(k_database=k_database)
        self.k_database=self.gas_mix.k_database
        self.Ng=self.gas_mix.Ng
        # to know whether we are dealing with corr-k or not and access some attributes. 
        if self.k_database is not None:
            #self.Nw = self.k_database.Nw
            #self.wnedges = self.k_database.wnedges
            #self.wns = self.k_database.wns
            #self.dwnedges = np.diff(self.wnedges)
            self.compute_albedo_surf_nu = True
            self.number_computation_albedo_surf_nu = 0
            self.compute_flux_top_dw_nu = True
            self.number_computation_flux_top_dw_nu = 0
            self.known_spectral_axis = False

    def set_cia_database(self, cia_database=None):
        """Change the CIA database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_cia_database` for details.

        Parameters
        ----------
            cia_database: :class:`CIAdatabase` object
                New CIAdatabase to use.
        """
        self.gas_mix.set_cia_database(cia_database=cia_database)

    def set_a_database(self, a_database=None):
        """Change the Aerosol database used by the
        :class:`Aerosols` object handling aerosol optical properties.

        See :any:`aerosols.Aerosols.set_a_database` for details.

        Parameters
        ----------
            a_database: :class:`Adatabase` object
                New Adatabase to use.
        """    
        self.aerosols.set_a_database(a_database=a_database)

    def set_spectral_range(self, wn_range = None, wl_range = None):
        """Sets the spectral range in which computations will be done by specifying
        either the wavenumber (in cm^-1) or the wavelength (in micron) range.

        See :any:`gas_mix.Gas_mix.set_spectral_range` for details.
        """
        self.gas_mix.set_spectral_range(wn_range = wn_range, wl_range = wl_range)
        self.compute_albedo_surf_nu = True
        self.compute_flux_top_dw_nu = True
        self.known_spectral_axis = False

    def set_incoming_stellar_flux(self, flux_top_dw: Optional[float] = None, Tstar: Optional[float] = None,
                                  stellar_spectrum: Optional[Spectrum] = None, **kwargs):
        """Sets the stellar incoming flux integrated in each wavenumber
        channel.

        .. important::
            The normalization is such that the flux input is exactly
            flux_top_dw whatever the spectral range used.

        Parameters
        ----------
            flux_top_dw:
                Bolometric Incoming flux (in W/m^2).
            Tstar:
                Stellar temperature (in K) used to compute the spectral distribution
                of incoming flux using a blackbody.
            stellar_spectrum:
                Spectrum of the star in units of per wavenumber. The specific units do not matter as
                the overall flux will be renormalized.
        """
        if flux_top_dw is not None:
            self.flux_top_dw = flux_top_dw
            self.compute_flux_top_dw_nu = True
        if (Tstar is not None) and (stellar_spectrum is not None):
            print('Carefull, both Tstar and stellar_spectrum in set_incoming_stellar_flux')
            print('Tstar will not be used.')
        if Tstar is not None:
            self.Tstar = Tstar
            self.stellar_spectrum = None
            self.compute_flux_top_dw_nu = True
        if stellar_spectrum is not None:
            self.stellar_spectrum = stellar_spectrum
            self.Tstar = None
            self.compute_flux_top_dw_nu = True

        if (self.compute_flux_top_dw_nu) and (self.known_spectral_axis):
            if self.stellar_spectrum is not None:
                binned_spec = self.stellar_spectrum.bin_down_cp(self.wnedges)
                self.flux_top_dw_nu = binned_spec.value
                factor = self.flux_top_dw / (binned_spec.total)
            elif self.Tstar is not None:
                self.flux_top_dw_nu = Bnu_integral_num(self.wnedges, self.Tstar)
                factor = self.flux_top_dw / (np.sum(self.flux_top_dw_nu) * self.dwnedges)
            else:
                raise RuntimeError('Something went wrong, Tstar and stellar_spectrum cannot be both None')
            self.flux_top_dw_nu = self.flux_top_dw_nu * factor
            self.compute_flux_top_dw_nu = False
            self.number_computation_flux_top_dw_nu += 1
    
    def set_internal_flux(self, internal_flux):
        """Sets internal flux from the subsurface in W/m^2
        """
        self.internal_flux = internal_flux

    def set_surface_albedo(self, albedo_surf: Optional[float] = None, wn_albedo_cutoff: Optional[float] = None,
                           **kwargs):
        """Sets the value of the mean surface albedo.
        
        Parameters
        ----------
            albedo_surf: float
                Effective visible surface albedo.

            wn_albedo_cutoff : float
                wavenumber value in cm-1 dividing the visible range from the IR range, 
                where the albedo goes from a given value 'albedo_surf' to 0.
        """
        if albedo_surf is not None:
            self.albedo_surf = albedo_surf
            self.compute_albedo_surf_nu = True
        if wn_albedo_cutoff is not None:
            self.wn_albedo_cutoff = wn_albedo_cutoff
            self.compute_albedo_surf_nu = True
        if ((self.compute_albedo_surf_nu) and (self.known_spectral_axis)):
            self._setup_albedo_surf_nu()
            self.compute_albedo_surf_nu = False
            self.number_computation_albedo_surf_nu += 1

    def _setup_albedo_surf_nu(self):
        """Compute the value of the mean surface albedo for each wavenumber in an array.
        If the albedo value is modified, this method needs to be called again.
        For now, only available with wavenumbers
        """
        self.albedo_surf_nu = np.where(self.wns >= self.wn_albedo_cutoff, self.albedo_surf, 0.)
    
    def set_rayleigh(self, rayleigh: bool = False):
        """Sets whether or not Rayleigh scattering is included."""
        self.rayleigh = rayleigh

    def spectral_integration(self, spectral_var):
        """Spectrally integrate an array, taking care of whether
        we are dealing with corr-k or xsec data.

        Parameters
        ----------
            spectral_var: array, np.ndarray
                array to integrate

        Returns
        -------
            var: array, np.ndarray
                array integrated over wavenumber (and g-space if relevant)
        """

        return np.sum(self.g_integration(spectral_var) * self.dwnedges, axis=-1)

    def g_integration(self, spectral_var):
        """Integrate an array along the g direction (only for corrk)

        Parameters
        ----------
            spectral_var: array, np.ndarray
                array to integrate

        Returns
        -------
            var: array, np.ndarray
                array integrated over g-space if relevant
        """
        if self.Ng is None:
            var = spectral_var
        else:
            var = np.sum(spectral_var * self.k_database.weights, axis=-1)
        return var

    def opacity(self, rayleigh = None, compute_all_opt_prop = False,
            wn_range = None, wl_range = None, **kwargs):
        """Computes the opacity of each of the radiative layers (m^2/molecule).

        Parameters
        ----------
            rayleigh: bool
                If true, the rayleigh cross section is computed in
                self.kdata_scat and added to kdata(total extinction cross section)
                If None, the global attribute self.rayleigh is used.

        See :any:`gas_mix.Gas_mix.cross_section` for details.
        """
        if rayleigh is None:
            local_rayleigh = self.rayleigh
        else:
            local_rayleigh = rayleigh

        self.kdata = self.gas_mix.cross_section(rayleigh=local_rayleigh,
                                                wn_range=wn_range, wl_range=wl_range, **kwargs)
        shape = self.kdata.shape

        if ((wn_range is not None) or (wl_range is not None)):
            self.compute_albedo_surf_nu = True
            self.compute_flux_top_dw_nu = True
            self.known_spectral_axis = False

        if self.aerosols.adatabase is not None:
            [kdata_aer, k_scat_aer, g_aer] = self.aerosols.optical_properties(
                    compute_all_opt_prop=compute_all_opt_prop, **kwargs)
            if self.Ng is None:
                self.kdata += kdata_aer
            else:
                self.kdata += kdata_aer[:,:,None]

        if compute_all_opt_prop:
            if local_rayleigh:
                kdata_scat_tot = self.gas_mix.kdata_scat
            else:
                kdata_scat_tot = np.zeros(shape[0:2], dtype=float)

            if self.aerosols.adatabase is not None:
                kdata_scat_tot += k_scat_aer
                self.asym_param = np.where(kdata_scat_tot<=0., 0., k_scat_aer * g_aer / kdata_scat_tot)
                #the line below is for test. Should be removed
                #self.asym_param = np.zeros(shape[0:2], dtype=float)
                #self.asym_param = np.ones(shape[0:2], dtype=float)
            else:
                self.asym_param = np.zeros(shape[0:2], dtype=float)

            if self.Ng is None:
                self.single_scat_albedo = kdata_scat_tot / self.kdata
            else:
                self.single_scat_albedo = kdata_scat_tot[:,:,None] / self.kdata
                self.asym_param = self.asym_param[:,:,None] * np.ones(self.Ng)
                #JL21 the line above could be removed as asym param does not seem to change
                # with g point, but then the dimensions should be change in 2stream routines.

            self.single_scat_albedo=np.core.umath.minimum(self.single_scat_albedo,0.9999999999)
            #self.single_scat_albedo=np.core.umath.maximum(self.single_scat_albedo,0.1)

        self.Nw=self.gas_mix.Nw
        self.wns=self.gas_mix.wns
        self.wnedges=self.gas_mix.wnedges
        self.dwnedges=self.gas_mix.dwnedges
        self.known_spectral_axis = True

    def source_function(self, integral=True, source=True):
        """Compute the blackbody source function (Pi*Bnu) for each layer of the atmosphere.

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
            source: boolean, optional
                If False, the source function is put to 0 (for solar absorption calculations)
        """
        if source:
            if integral:
                #JL2020 What is below is slightly faster for very large resolutions
                #piBatm=np.empty((self.Nlay,self.Nw))
                #for ii in range(self.Nlay):
                #    piBatm[ii]=PI*Bnu_integral_num(self.wnedges,self.tlay[ii])/dw
                #JL2020 What is below is much faster for moderate to low resolutions
                piBatm=PI*Bnu_integral_array(self.wnedges,self.tlay,self.Nw,self.Nlay) \
                    /self.dwnedges
            else:
                piBatm=PI*Bnu(self.wns[None,:],self.tlay[:,None])
        else:
            piBatm=np.zeros((self.Nlay,self.Nw))
        return piBatm

    def setup_emission_caculation(self, mu_eff=0.5, rayleigh=None, integral=True,
            source=True, gas_vmr=None, Mgas=None, aer_reffs_densities=None, **kwargs):
        """Computes all necessary quantities for emission calculations
        (opacity, source, etc.)
        """
        if gas_vmr is not None: self.set_gas(gas_vmr, Mgas = Mgas)
        if aer_reffs_densities is not None: self.aerosols.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)
        self.opacity(rayleigh = rayleigh, **kwargs)
        self.set_surface_albedo(**kwargs)
        self.set_incoming_stellar_flux(**kwargs)
        self.piBatm = self.source_function(integral=integral, source=source)
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau, self.dtau=rad_prop_xsec(self.dcol_density_rad,
                self.kdata, mu_eff)
        else:
            self.tau, self.dtau=rad_prop_corrk(self.dcol_density_rad,
                self.kdata, mu_eff)
            self.weights=self.k_database.weights

    def emission_spectrum(self, integral=True, mu0=0.5, mu_quad_order=None,
            dtau_min=1.e-13, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            mu_quad_order: int
                If an integer is given, the emission intensity is computed
                for a number of angles and integrated following a gauss legendre
                quadrature rule of order `mu_quad_order`.
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        if mu_quad_order is not None:
            # if we want quadrature, use the more general method.
            return self.emission_spectrum_quad(integral=integral,
                mu_quad_order=mu_quad_order, dtau_min=dtau_min, **kwargs)

        try:
            self.setup_emission_caculation(mu_eff=mu0, rayleigh = False, integral=integral, **kwargs)
        except TypeError:
            raise RuntimeError("""
            Cannot use rayleigh option with emission_spectrum.
            If you meant to include scattering, you should use emission_spectrum_2stream.
            """)
        # self.tau and self.dtau include the 1/mu0 factor.
        expdtau=np.exp(-self.dtau)
        expdtauminone=np.where(self.dtau < dtau_min, -self.dtau, expdtau-1.)
        # careful: due to numerical limitations, 
        # the limited development of Exp(-dtau)-1 needs to be used for small values of dtau
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            timesBatmTop=(1.+expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom=(-expdtau-expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom[-1]+=exptau[-1]
        else:
            timesBatmTop=np.sum((1.+expdtauminone/self.dtau)*exptau[:-1]*self.weights,axis=-1)
            timesBatmBottom=np.sum((-expdtau-expdtauminone/self.dtau)*exptau[:-1] \
                *self.weights,axis=-1)
            timesBatmBottom[-1]+=np.sum(exptau[-1]*self.weights,axis=-1)
        IpTop=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0)

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_quad(self, integral=True, mu_quad_order=3, dtau_min=1.e-13, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)
        using gauss legendre qudrature of order `mu_quad_order`

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.
                  
        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh = False, integral=integral, **kwargs)
        # angle effect dealt with later

        IpTop = np.zeros(self.kdata.shape[1])
        mu_w, mu_a, _ = gauss_legendre(mu_quad_order)
        mu_w = mu_w * mu_a * 2.  # takes care of the mu factor in last integral => int(mu I d mu)
        # Factor 2 takes care of the fact that the source function is pi*Batm
        # but we want 2*Pi*Batm
        for ii, mu0 in enumerate(mu_a):
            tau=self.tau/mu0
            dtau=self.dtau/mu0
            expdtau=np.exp(-dtau)
            expdtauminone=np.where(dtau<dtau_min,-dtau,expdtau-1.)
            exptau=np.exp(-tau)
            if self.Ng is None:
                timesBatmTop=(1.+expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom=(-expdtau-expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom[-1]+=exptau[-1]
            else:
                timesBatmTop=np.sum((1.+expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom=np.sum((-expdtau-expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom[-1]+=np.sum(exptau[-1]*self.weights,axis=-1)
            IpTop+=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0) \
                *mu_w[ii]

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_2stream(self, integral=True, mu0=0.5,
            method='toon', dtau_min=1.e-10, flux_at_level=False, rayleigh = None,
            flux_top_dw=None, source=True, compute_kernel=False, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If True, the source function (black body)
                  is integrated within each wavenumber bin.
                * If False, only the central value is used.
                  False is faster and should be ok for small bins
                  (e.g. used with `Xtable`),
                  but True is the most accurate mode. 
            rayleigh: bool
                Whether to account for rayleigh scattering or not. 
                If None, the global attribute self.rayleigh is used.

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.
            flux_at_level: bool
                Whether the fluxes are computed at the levels (True) or
                in the middle of the layers (False). e.g. this needs to 
                be True to be able to use the kernel with the evolution model.
            source: bool
                Whether to include the self emission of the gas in the computation.
            compute_kernel: bool
                Whether we want to recompute the kernel of the heating rates along
                with the fluxes.
            method: str
                What method to use to computed the 2 stream fluxes. 
                'toon' (default) uses Toon et al. (1989).
                'lmdz' is based on the same equation, but the implementation
                is closer to the actual method used in the LMDZ GCM. This mode
                is not supported anymore and might be broken!

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh = rayleigh, integral=integral,
            source=source, flux_top_dw=flux_top_dw, compute_all_opt_prop=True, **kwargs)
        # mu_eff=1. because the mu effect is taken into account in solve_2stream_nu
                 # we must compute the vertical optical depth here.
        # JL21: shouldn't we remove flux_top_dw, it seems not to be used.
        self.single_scat_albedo=np.where(self.dtau<dtau_min,0.,self.single_scat_albedo)
        self.dtau=np.where(self.dtau<dtau_min,dtau_min,self.dtau)

        module_to_use=globals()[method]
        # globals()[method] converts the method string into a module name
        #  if the module has been loaded
        if self.Ng is None:
            solve_2stream_nu=module_to_use.solve_2stream_nu_xsec
        else:
            solve_2stream_nu=module_to_use.solve_2stream_nu_corrk

        if flux_top_dw is not None:
            self.set_incoming_stellar_flux(flux=flux_top_dw, **kwargs)

        self.flux_up_nu, self.flux_down_nu, self.flux_net_nu = \
            solve_2stream_nu(self.piBatm, self.dtau, self.single_scat_albedo, self.asym_param,
                self.flux_top_dw_nu, self.albedo_surf_nu, mu0 = mu0, flux_at_level=flux_at_level)
        if compute_kernel: self.compute_kernel(solve_2stream_nu, mu0=mu0, flux_at_level=flux_at_level,
                    per_unit_mass=True, integral=True, **kwargs)

        if self.Ng is None:
            return Spectrum(self.flux_up_nu[0],self.wns,self.wnedges)
        else:
            return Spectrum(np.sum(self.flux_up_nu[0]*self.weights,axis=1),self.wns,self.wnedges)

    def compute_kernel(self, solve_2stream_nu, epsilon=0.01, flux_at_level=False, mu0 = 0.5,
            per_unit_mass=True, integral=True, **kwargs):
        """Computes the Jacobian matrix d Heating[lay=i] / d T[lay=j]

        This method should not be called by the user directly. The recompute the
        kernel, call emission_spectrum_2stream with the option compute_kernel=True.

        After the computation, the kernel is stored in self.kernel and the
        temperature array at wich the kernel has been computed in 
        self.tlay_kernel.

        Parameters
        ----------
            solve_2stream_nu: function
                Name of the function that will be used to compute the
                fluxes. e.g.: two_stream_toon or two_stream_lmdz 
            epsilon: float
                The temperature increment used to compute the derivative
                will be epsilon*self.tlay
            flux_at_level: bool
                Whether the fluxes are computed at the levels (True) or
                in the middle of the layers (False). e.g. this needs to 
                be True to be able to use the kernel with the evolution model.
            mu0: float
                Cosine of the effective zenith angle used in the flux computations.
            per_unit_mass: bool
                Whether the heating rates will be normalized per unit of mass (i.e.
                the kernel values will have units of W/kg/K).
                If False, kernel in W/layer/K.
            integral: bool
                Whether the integral mode is used in flux computations. e.g. this needs to 
                be True to be able to use the kernel with the evolution model. 
        """
        net=self.spectral_integration(self.flux_net_nu)
        self.kernel=np.empty((self.Nlay,self.Nlay))
        tlay=self.tlay
        dT = epsilon*tlay
        self.tlay = tlay + dT
        newpiBatm = self.source_function(integral=integral)

        for ilay in range(self.Nlay):
            pibatm = np.copy(self.piBatm)
            pibatm[ilay] = newpiBatm[ilay]
            _, _, flux_net_tmp = \
                solve_2stream_nu(pibatm, self.dtau, self.single_scat_albedo, self.asym_param,
                    self.flux_top_dw_nu, self.albedo_surf_nu, mu0 = mu0, flux_at_level=flux_at_level)
            net_tmp = self.spectral_integration(flux_net_tmp)
            self.kernel[ilay]=(net-net_tmp)/dT[ilay]
        self.kernel[:,:-1]-=self.kernel[:,1:]
        self.tlay=tlay
        self.tlay_kernel=self.tlay
        if per_unit_mass: self.kernel*=self.inv_dmass

    def flux_divergence(self, per_unit_mass = True, **kwargs):
        """Computes the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            H: array, np.ndarray
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
            net: array, np.ndarray
                Net fluxes at level surfaces
        """
        if self.flux_net_nu is None:
            raise RuntimeError('should have ran emission_spectrum_2stream.')
        net = self.spectral_integration(self.flux_net_nu)
        H = -np.copy(net)
        H[:-1] -= H[1:]
        H[-1] += self.internal_flux
        if per_unit_mass: H *= self.inv_dmass
        return H, net

    def heating_rate(self, compute_kernel=False, dTmax_use_kernel=None, **kwargs):
        """Computes the heating rates and net fluxes in the atmosphere.

        Parameters
        ----------
            compute_kernel: bool
                If True, the kernel (jacobian of the heat rates) is recomputed
            dTmax_use_kernel: float
                Maximum temperature difference between the current temperature
                and the temperature of the last kernel computation before
                a new kernel is recomputed.

        Returns
        -------
            H: array, np.ndarray
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
            net: array, np.ndarray
                Net fluxes at level surfaces
        """
        if (not compute_kernel) and (dTmax_use_kernel is not None):
            dT=self.tlay-self.tlay_kernel
            if np.amax(np.abs(dT)) < dTmax_use_kernel:
                try:
                    H = self.H_kernel + np.dot(dT,self.kernel)
                    net = self.internal_flux - np.cumsum((H*self.dmass)[::-1])[::-1]
                except:
                    raise RuntimeError("Kernel has not been precomputed")
                return H, net
        _ = self.emission_spectrum_2stream(flux_at_level=True, integral=True,
                compute_kernel=compute_kernel, **kwargs)
        H, net = self.flux_divergence(compute_kernel=compute_kernel, **kwargs)
        if compute_kernel:
            self.H_kernel = H
            self.tau_rads = 1./np.abs(self.kernel.diagonal())
            self.tau_rad = np.amin(self.tau_rads)
        return H, net

    def bolometric_fluxes_2band(self, flux_top_dw = None, **kwargs):
        """Computes the bolometric fluxes at levels and heating rates using `bolometric_fluxes`.

        However, the (up, down, net) fluxes are separated in two contributions:
          - the part emitted directly by the atmosphere (_emis).
          - the part due to the incoming stellar light (_stell),
            that can be used to compute the absorbed stellar radiation and the bond_albedo.

        We also provide the associated heating rates (H in W/kg)
        """
        save_stellar_flux = np.copy(self.flux_top_dw_nu)
        self.flux_top_dw_nu = self.flux_top_dw_nu * 0.
        _ = self.emission_spectrum_2stream(flux_at_level = True, integral = True, **kwargs)
        Fup_emis, Fdw_emis, Fnet_emis, H_emis = self.bolometric_fluxes(**kwargs)
        self.flux_top_dw_nu = save_stellar_flux

        save_internal_flux = np.copy(self.internal_flux)
        _ = self.emission_spectrum_2stream(flux_at_level = True, integral = True,
        flux_top_dw = flux_top_dw, source = False, **kwargs)
        Fup_stel, Fdw_stel, Fnet_stel, H_stel = self.bolometric_fluxes(**kwargs)
        self.internal_flux = save_internal_flux

        return Fup_emis, Fdw_emis, Fnet_emis, H_emis, Fup_stel, Fdw_stel, Fnet_stel, H_stel

    def spectral_fluxes_2band(self, **kwargs):
        """Computes the spectral fluxes at levels.

        The (up, down, net) fluxes are separated in two contributions:
          - the part emitted directly by the atmosphere (_emis).
          - the part due to the incoming stellar light (_stell),
            that can be used to compute the absorbed stellar radiation and the bond_albedo.
        """
        save_stellar_flux = np.copy(self.flux_top_dw_nu)
        self.flux_top_dw_nu = self.flux_top_dw_nu * 0.
        _ = self.emission_spectrum_2stream(flux_at_level=True, integral=True, **kwargs)
        Fup_emis = self.g_integration(self.flux_up_nu)
        Fdw_emis = self.g_integration(self.flux_down_nu)
        Fnet_emis = self.g_integration(self.flux_net_nu)
        self.flux_top_dw_nu = save_stellar_flux

        _ = self.emission_spectrum_2stream(flux_at_level=True, integral=True, source=False, **kwargs)
        Fup_stel = self.g_integration(self.flux_up_nu)
        Fdw_stel = self.g_integration(self.flux_down_nu)
        Fnet_stel = self.g_integration(self.flux_net_nu)

        return Fup_emis, Fdw_emis, Fnet_emis, Fup_stel, Fdw_stel, Fnet_stel

    def bolometric_fluxes(self, per_unit_mass = True, **kwargs):
        """Computes the bolometric fluxes at levels and the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            up: array, np.ndarray
                Upward fluxes at level surfaces
            dw: array, np.ndarray
                Downward fluxes at level surfaces
            net: array, np.ndarray
                Net fluxes at level surfaces
            H: array, np.ndarray
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
        """
        H, net = self.flux_divergence(per_unit_mass = per_unit_mass)
        up=self.spectral_integration(self.flux_up_nu)
        dw=self.spectral_integration(self.flux_down_nu)
        return up, dw, net, H

    def transmittance_profile(self, **kwargs):
        """Computes the transmittance profile of an atmosphere,
        i.e. Exp(-tau) for each layer of the model.
        Real work done in the numbafied function path_integral_corrk/xsec
        depending on the type of data.
        """
        self.opacity(**kwargs)
        self.compute_tangent_path()
        self.compute_number_density()
        if self.Ng is not None:
            self.weights=self.k_database.weights
            transmittance=path_integral_corrk( \
                self.Nlay-1,self.Nw,self.Ng,self.tangent_path,self.n_density,self.kdata,self.weights)
        else:
            transmittance=path_integral_xsec( \
                self.Nlay-1,self.Nw,self.tangent_path,self.n_density,self.kdata)
        return transmittance

    def transmission_spectrum(self, normalized=False, Rstar=None, **kwargs):
        r"""Computes the transmission spectrum of the atmosphere.
        In general (see options below), the code returns the transit depth:

        .. math::
            \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2),

        where

        .. math::
          \alpha_\nu=2 \pi \int_0^{z_{max}} (R_p+z)*(1-e^{-\tau_\nu(z)) d z.
        
        Parameters
        ----------
            Rstar: float or astropy.unit object, optional
                Radius of the host star. If a float is specified, meters are assumed.
                Does not need to be given here if
                it has already been specified as an attribute of the :class:`Atm` object.
                If specified, the result is the transit depth:

                .. math::
                  \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2).

            normalized: boolean, optional
                Used only if self.Rstar and Rstar are None:

                * If True,
                  the result is normalized to the planetary radius:

                  .. math::
                    \delta_\nu=1+\frac{\alpha_\nu}{\pi R_p^2}.
                * If False,
                                    
                  .. math::
                    \delta_\nu=\pi R_p^2+\alpha_\nu.

        Returns
        -------
            array
                The transit spectrum (see above for normalization options).
        """
        if Rstar is not None: self.set_Rstar(Rstar)
        transmittance=self.transmittance_profile(**kwargs)
        self.compute_area()
        res=Spectrum((np.dot(self.area,(1.-transmittance))),self.wns,self.wnedges)
        if self.Rstar is not None:
            return (res+(PI*self.Rp**2))/(PI*self.Rstar**2)
        elif normalized:
            return res/(PI*self.Rp**2)+1
        else:
            return res+(PI*self.Rp**2)

    def __repr__(self):
        """Method to output header
        """
        output=super().__repr__()
        output+="""
    k_database      :
        {kdatab}
    cia_database    :
        {cdatab}""".format(kdatab=self.k_database, cdatab=self.gas_mix.cia_database)
        if self.gas_mix._wn_range is not None:
            output+='    wn range        : '+ self.gas_mix._wn_range +'\n'

        return output
        
    def exp_minus_tau(self):
        """Sums Exp(-tau) over gauss points
        """
        weights=self.k_database.weights
        return np.sum(np.exp(-self.tau[1:])*weights,axis=2)

    def exp_minus_tau_g(self, g_index):
        """Sums Exp(-tau) over gauss point
        """
        return np.exp(-self.tau[1:,:,g_index])

    def blackbody(self, layer_idx=-1, integral=True):
        """Computes the surface black body flux (in W/m^2/cm^-1) for the temperature
        of layer `layer_idx`.

        Parameters
        ----------
            layer_idx; int
                Index of layer used for the temperature.
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux in W/m^2/cm^-1
        """
        if not self.known_spectral_axis:
            self.opacity(integral = integral)
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlay[layer_idx])/self.dwnedges
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlay[layer_idx])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def surf_bb(self, **kwargs):
        """Computes the surface black body flux (in W/m^2/cm^-1).

        See :any:`blackbody` for options.

        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the surface (in W/m^2/cm^-1)
        """
        return self.blackbody(layer_idx=-1, **kwargs)

    def top_bb(self, **kwargs):
        """Computes the top of atmosphere black body flux (in W/m^2/cm^-1).

        See :any:`blackbody` for options.

        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the temperature at the top of atmosphere (in W/m^2/cm^-1)
        """
        return self.blackbody(layer_idx=0, **kwargs)

    def contribution_function(self) -> np.ndarray:
        r"""
        Compute the contribution function $Cf(P,\lambda) = B(T,\lambda) \frac{d e^\tau}{d \log P}$.

        $\tau$ and $\log P$ are taken at the mid-layers, the temperature needs to be taken at the level surfaces.
        For that, we use the computed temperature `t_opac[:-1]`.

        The result is not normalized to allow comparison.
        Returns
        -------
        np.ndarray
            Contribution function [W/m^2/str/cm-1].
            Shape: **(Nlay - 1, Nw)**.

        """
        B = Bnu_integral_array(self.wnedges, self.t_opac, self.Nw,
                               self.t_opac.shape[0]) / self.dwnedges
        dlogP = np.diff(self.logplay)
        detau = np.diff(self.g_integration(np.exp(-self.tau)), axis=0)

        return - ((B * detau).T / dlogP).T



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A module to handle ouputs rebinning and plotting
"""
from __future__ import annotations

from typing import Optional

import astropy.units as u
import h5py
import numpy as np

from exo_k.util.interp import rebin
from exo_k.util.spectral_object import Spectral_object
from exo_k.util.cst import PLANCK, C_LUM

planckovclum = PLANCK * 100. * C_LUM  #100 is a conversion factor from cm^-1 to m^-1

class Spectrum(Spectral_object):
    """A class defining a Spectrum object to plot and manipulate.
    """

    def __init__(self, value = None, wns = None, wnedges = None, input_spectral_unit='cm^-1', spectral_unit='cm^-1',
            filename = None, spectral_radiance = False, 
            from_taurex = False, dataset = 'native_spectrum', **kwargs):
        """Instanciate with a value, bin centers, and bin edges.
        Can also load a Taurex spectrum if filename is provided.

        Parameters
        ----------
            value: Array
                spectrum values
            wns: Array
                Spectral grid (can be wavenumbers or wavelengths)
            wnedges: Array
                Bin edges grid (can be wavenumbers or wavelengths)
            input_spectral_unit: str
                Unit of the input spectral grid
            spectral_unit: str
                desired output unit for the spectral grid
            filename: str
                Name of a file where the spectrum is stored. 
                If a filename is given, there is no need to provide
                values, wns, ...
            spectral_radiance: bool
                If True, the spectrum is assumed to be a flux in units of
                inverse spectral units (for example W/micron if the spectral unit is microns)
                If False, the spectrum is considered as monochromatic values (like Rp/Rs**2)
        """
        super().__init__()
        self.value = value
        self.wns = wns
        self.wnedges = wnedges
        self.spec_unit = input_spectral_unit
        self.spectral_radiance = spectral_radiance

        if filename is not None:
            if from_taurex:
                self.load_taurex(filename, dataset)
            elif filename.lower().endswith(('.hdf5', '.h5')):
                self.read_hdf5(filename)
            elif filename.lower().endswith(('.dat', '.txt')):
                self.read_ascii(filename, **kwargs)

        if (self.wnedges is None) and (self.wns is not None):
            self.wnedges = np.concatenate(([self.wns[0]],
                (self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))

        self.convert_spectral_units(input_spectral_unit=input_spectral_unit,
                                    spectral_unit=spectral_unit, spectral_radiance=spectral_radiance)

        if self.wnedges is not None:
            self.dwnedges = np.diff(self.wnedges)
        else:
            self.dwnedges = None

    def convert_spectral_units(self, input_spectral_unit='cm^-1', spectral_unit='cm^-1',
                               spectral_radiance: Optional[bool] = None):
        if spectral_unit != input_spectral_unit:
            wns_tmp = (self.wns*u.Unit(input_spectral_unit)).to(u.Unit(spectral_unit), equivalencies=u.spectral()).value
            wnedges_tmp = (self.wnedges*u.Unit(input_spectral_unit)).to(u.Unit(spectral_unit), equivalencies=u.spectral()).value

            if spectral_radiance is not None:
                self.spectral_radiance = spectral_radiance

            if self.spectral_radiance:
                dwnedges = np.diff(self.wnedges)
                dwnedges_tmp = np.diff(wnedges_tmp)
                self.value = self.value * np.abs(dwnedges / dwnedges_tmp)
                # assumes that the wavelength unit
                # is the same as the spectral unit (for example F in W/nm if wl in nm).

            if wnedges_tmp[-1] > wnedges_tmp[0]:
                self.wns = wns_tmp
                self.wnedges = wnedges_tmp
            else:
                self.wns = np.copy(wns_tmp[::-1])
                self.wnedges = np.copy(wnedges_tmp[::-1])
                self.value = np.copy(self.value[::-1])

    def normalize(self, bolometric_flux: float):
        """Normalize the spectrum to a specified bolometric flux

        Parameters
        ----------
            bolometric_flux: float
                Integral of the flux over the total bandpass after normalization. 
        """
        factor = bolometric_flux / self.total
        self.value *= factor

    def photonize(self):
        """Translate to photon count in each bin
        """
        self.value /= self.wns * planckovclum

    def dephotonize(self):
        """Translate back to energy
        """
        self.value *= self.wns * planckovclum

    def integrate_per_bin(self):
        """integrate energy in each bin
        """
        self.value *= self.dwnedges

    def copy(self):
        """Deep copy of the spectrum.
        """
        return Spectrum(self.value.copy(), self.wns.copy(), self.wnedges.copy(),
                        input_spectral_unit=self.spec_unit, spectral_unit=self.spec_unit)

    def plot_spectrum(self, ax, per_wavenumber=True, x_axis='wls',
                      xscale=None, yscale=None, **kwarg):
        """Plot the spectrum
        
        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            per_wavenumber: bool, optional
                Defines the units of spectral flux density.
                False converts to per wavelength units.
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            xscale, yscale: str, optional
                If 'log' log axes are used.
        """
        if per_wavenumber:
            toplot=self.value
        else:
            toplot=self.value/self.wls**2*1.e4
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.set_ylabel('Flux')
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def bin_down(self, wnedges):
        """Bins down the spectrum to a new grid of wnedges by conserving area.
        
        Parameters
        ----------
            wnedges: array, np.ndarray
                Wavenumbers of the bin edges to be used
        """
        wnedges=np.array(wnedges)
        self.value=rebin(self.value,self.wnedges,wnedges)
        self.wnedges=wnedges
        self.dwnedges = np.diff(self.wnedges)
        self.wns=0.5*(self.wnedges[:-1]+self.wnedges[1:])

    def bin_down_cp(self, wnedges):
        """Returns a new binned down spectrum to a grid of wnedges by conserving area.
        
        Parameters
        ----------
            wnedges: array, np.ndarray
                Wavenumbers of the bin edges to be used

        Returns
        -------
            :class:`Spectrum`
                Binned down spectrum
        """
        res=self.copy()
        res.bin_down(wnedges)
        return res

    def clip_spectral_range(self, wn_range=None, wl_range=None):
        """Limits the data to the provided spectral range (inplace):

           * Wavenumber in cm^-1 if using wn_range argument
           * Wavelength in micron if using wl_range
        """
        iw_min, iw_max = self.select_spectral_range(wn_range, wl_range)
        self.value = self.value[iw_min:iw_max]

    def randomize(self, uncertainty = 0.):
        """Adds random noise with a given uncertainty.
        """
        self.noise = uncertainty * np.random.randn((self.value.size))
        self.value += self.noise

    def __add__(self, other):
        """Defines addition
        """
        if isinstance(other, (float, int, np.ndarray)):
            return Spectrum(self.value + other, self.wns, self.wnedges)

        if (self.wns.size == other.wns.size) and np.array_equal(self.wns, other.wns):
            val = self.value + other.value
            return Spectrum(val, self.wns, self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __radd__(self, other):
        """Use commutativity v + S == S + v"""

        return self.__add__(other)

    def __sub__(self, other):
        """Defines subtraction
        """
        if isinstance(other, (float, int, np.ndarray)):
            return Spectrum(self.value - other, self.wns, self.wnedges)

        if (self.wns.size == other.wns.size) and np.array_equal(self.wns, other.wns):
            val = self.value - other.value
            return Spectrum(val, self.wns, self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __mul__(self, other):
        """Defines multiplication
        """
        if isinstance(other, (float, int, np.ndarray)):
            return Spectrum(self.value * other, self.wns, self.wnedges)

        if (self.wns.size == other.wns.size) and np.array_equal(self.wns, other.wns):
            val = self.value * other.value
            return Spectrum(val, self.wns, self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __rmul__(self, other):
        """Use commutativity v * S == S * v"""

        return self.__mul__(other)

    def __truediv__(self, other):
        """Defines division
        """
        if isinstance(other, (float, int, np.ndarray)):
            return Spectrum(self.value / other, self.wns, self.wnedges)

        if (self.wns.size == other.wns.size) and np.array_equal(self.wns, other.wns):
            val = self.value / other.value
            return Spectrum(val, self.wns, self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __rtruediv__(self, other):
        """Defines division
        """
        if isinstance(other, (float, int, np.ndarray)):
            return Spectrum(other / self.value, self.wns, self.wnedges)

        if (self.wns.size == other.wns.size) and np.array_equal(self.wns, other.wns):
            val = other.value / self.value
            return Spectrum(val, self.wns, self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def std(self):
        """Defines standard deviation
        """
        return self.value.std()

    def abs(self):
        """Defines absolute value
        """
        return Spectrum(np.abs(self.value),self.wns,self.wnedges)

    def log10(self):
        """Defines Log 10
        """
        return Spectrum(np.log10(self.value),self.wns,self.wnedges)

    @property
    def total(self):
        """Defines the weighted sum over the spectrum
        """
        return np.dot(self.value,self.dwnedges)

    def read_hdf5(self, filename=None):
        """Reads data in a hdf5 file

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        with h5py.File(filename, 'r') as f:
            self.wns=f['bin_centers'][...]
            self.wnedges=f['bin_edges'][...]
            if 'units' in f['bin_edges'].attrs:
                self.spec_unit=f['bin_edges'].attrs['units']
            else:
                if 'units' in f['bin_centers'].attrs:
                    self.spec_unit=f['bin_centers'].attrs['units']
            self.value=f['spectrum'][...]

    def write_hdf5(self, filename):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.h5'
        compression="gzip"
        with h5py.File(fullfilename, 'w') as f:
            f.create_dataset("spectrum", data=self.value, compression=compression)
            f.create_dataset("bin_edges", data=self.wnedges, compression=compression)
            f["bin_edges"].attrs["units"] = 'cm^-1'
            f.create_dataset("bin_centers", data=self.wns, compression=compression)
            f["bin_centers"].attrs["units"] = 'cm^-1'

    def read_ascii(self, filename, usecols = (0,1), skip_header=0):
        """Saves data in a ascii format

        Parameters
        ----------
            filename: str
                Name of the file to be read
            spec_axis: str
                Whether the spectral axis in the file is
                wavenumber in cm^-1 ('wns') or wavelength in microns ('wls')
            skip_header: int
                Number of lines to skip
        """
        raw=np.genfromtxt(filename, skip_header = skip_header,
            usecols = usecols, names=('spec_axis','value'))
        self.wns = raw['spec_axis']
        self.value = raw['value']

    def write_ascii(self, filename, fmt='%.18e', spec_axis='wns', header=None):
        """Saves data in a ascii format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.dat', '.txt')):
            fullfilename=filename+'.dat'
        head=header
        if spec_axis=='wns':
            if head is None: head='wavenumber(cm^-1)     spectrum'
            np.savetxt(fullfilename, np.array([self.wns,self.value]).transpose(),
                fmt=fmt, header=head)
        else:    
            if head is None: head='wavelength(micron)    spectrum'
            np.savetxt(fullfilename, np.array([self.wls[::-1],self.value[::-1]]).transpose(),
                fmt=fmt, header=head)

    def load_taurex(self, filename,dataset='native_spectrum'):
        """Loads a taurex file

        Parameters
        ----------
            filename: str
                Full name (path) of the hdf5 file to load
            dataset: str
                Name of the hdf5 dataset to load
        """
        with h5py.File(filename, 'r') as f:
            self.wns = f['Output/Spectra/native_wngrid'][...]
            self.value = f['Output/Spectra/'+dataset][...]

        self.wnedges=np.concatenate(([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))

    def __repr__(self):
        """Method to output header
        """
        output="""
        value        : {val}
        wl (microns) : {wl}
        """.format(val=self.value,wl=self.wls)
        return output




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
from exo_k.util.singleton import Singleton
from .molar_mass import Molar_mass
from .cst import RGP

class Molar_heat_capacity(Singleton):
    """A class to compute the molar heat capacity at constant pressure
    (in J/mol/K) of regular molecules.
    This class can also store the molar mass of custom gases with arbitrary names
    (for example: My_gas, earth_background).
    """
    
    def init(self, *args, **kwds):
        """Initializes empty dictionary of custom heat capacities masses.
        """
        self._custom_molar_cp={}

    def add_species(self, species_dict):
        """Add one or several species to the database. 

        Parameters
        ----------
            species_dict: dict
                Keys are gases names (they do not have to be real molecules).
                Values are molar cp.
        """
        self._custom_molar_cp.update(species_dict)

    def __repr__(self):
        """Print the currently known species in the database. 
        """
        return self._custom_molar_cp.__repr__()

    def fetch(self, molecule_name):
        """yields the molar cp of a molecule
        
        Parameters:
            molecule_name: str
                Name of the molecule.

        Returns:
            float:
                Molar cp in J/mol/K
        """
        if molecule_name in self._custom_molar_cp.keys():
            return self._custom_molar_cp[molecule_name]
        if molecule_name in _specific_cp_at_300k.keys():
            Mmol = Molar_mass().fetch(molecule_name)
            molar_cp = _specific_cp_at_300k[molecule_name] * Mmol
            self._custom_molar_cp[molecule_name] = molar_cp
            return molar_cp
        else: # if we do not know, assume monoatomic gas
            return 1.5*RGP

# standard data at 300K
_specific_cp_at_300k = { \
'H2': 14310.,
'CO2': 744.,
'N2': 1040.,
'He': 5193.,
'H2O': 1864.,
'SO2': 640.,
'H2S': 1003.,
'CH4': 2226.,
'NH3': 2175,
'CO': 1040.,
'HCN': 1328.,
'O2': 918.,
'C2H6': 1763.,
'C2H2': 1575.,
}


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
import numba
from .convection import dry_convective_adjustment_numba, turbulent_diffusion_numba
from exo_k.util.molar_mass import Molar_mass
from exo_k.util.cst import PI, N_A
from exo_k.aerosol.util_aerosol import mmr_to_number_density_ratio

class Tracers(object):

    def __init__(self, settings, tracers=None, tracer_values=None, Kzz=0., Dmol=0.,
            bg_vmr=None, M_bg=None, Nlay=None, **kwargs):
        """Class that deals with tracers. 

        Fills out the tracers.qarray and creates tracer_names, a table of correspondence
        between tracer names and indices in tracers.qarray.
        """
        self.settings=settings
        if tracers is None:
            self.Ntrac=1
            tracers = {'inactive_tracer':{}}
        else:
            self.Ntrac = len(tracers)
        if tracer_values is None:
            tracer_values = {}
        if Nlay is not None:
            self.Nlay = Nlay
        else:
            raise RuntimeError("We need to know Nlay to initialize Tracers")
        self.bg_vmr = bg_vmr.copy()
        self.gas_vmr = bg_vmr.copy()
        self.var_gas_idx = list()
        self.aerosols_idx = list()
        self.var_gas_names = list()
        self.aerosols_names = list()
        self.gas_molar_masses = list()
        self.aerosols_bulk_density = list()
        self.aerosols_r_eff = list()
        self.some_var_gases = False
        self.some_active_aerosol = False
        self.dico=tracers
        self.namelist = list(tracers.keys())
        self.idx = dict()
        self.qarray = np.empty((self.Ntrac, self.Nlay))
        self.qsurf = np.zeros(self.Ntrac)
        self.qdeep = - np.ones(self.Ntrac)
        for ii, name in enumerate(self.namelist):
            self.idx[name]=ii
            if name in tracer_values.keys():
                self.qarray[ii]=np.copy(tracer_values[name])
            elif 'init_value' in self.dico[name].keys():
                self.qarray[ii]=np.ones(self.Nlay)*self.dico[name]['init_value']
            else:
                self.qarray[ii]=np.zeros(self.Nlay)
            if 'type' in self.dico[name]:
                if self.dico[name]['type'] in ('gas', 'vapor'):
                    self.some_var_gases = True
                    self.var_gas_idx.append(ii)
                    self.var_gas_names.append(name)
                    self.gas_molar_masses.append(Molar_mass().fetch(name))
                if self.dico[name]['type'] in ('aerosol'):
                    self.some_active_aerosol = True
                    self.aerosols_idx.append(ii)
                    self.aerosols_names.append(name)
                    self.aerosols_bulk_density.append(self.dico[name]['bulk_density'])
                    self.aerosols_r_eff.append(self.dico[name]['r_eff'])
                if 'q_deep' in self.dico[name]:
                    self.qdeep[ii] = np.copy(self.dico[name]['q_deep'])
            if 'surface_reservoir' in self.dico[name]: # should be reserved to condensates
                self.qsurf[ii] = np.copy(self.dico[name]['surface_reservoir'])
                self.settings.set_parameters(surface_reservoir = True)
        self.var_gas_idx = np.array(self.var_gas_idx)
        self.var_gas_names = np.array(self.var_gas_names)
        self.gas_molar_masses = np.array(self.gas_molar_masses)
        self.aerosols_idx = np.array(self.aerosols_idx)
        self.aerosols_names = np.array(self.aerosols_names)
        self.aerosols_bulk_density = np.array(self.aerosols_bulk_density)
        self.aerosols_r_eff = np.array(self.aerosols_r_eff)
        self.M_bg = M_bg
        self.update_gas_composition()
        self.Kzz = np.ones(self.Nlay)*Kzz
        self.Dmol = Dmol

    def update_gas_composition(self, update_vmr=True):
        """Performs mass to volume mixing ratio conversion,
        computes the new molar mass of the total gas
        and transmits new composition to the radiative model. 

        !!!Small discrepancies with case without tracers!!!
        Maybe a problem with the background gas. JL22 => is this still true ? 

        Parameters
        ----------
            atm: exo_k.Atm object
                If atm is provided, its composition will be
                updated with the new composition.
        """
        qvar = np.zeros(self.Nlay)
        ovMvar = np.zeros(self.Nlay)
        for ii, idx in enumerate(self.var_gas_idx):
            qvar += self.qarray[idx]
            ovMvar += self.qarray[idx]/self.gas_molar_masses[ii]
        ovMbg=1./self.M_bg
        self.Mgas = 1./((1.-qvar)*ovMbg + ovMvar)
        if update_vmr:
            var_vmr_tot = 0.
            for ii, idx in enumerate(self.var_gas_idx):
                vmr_tmp = np.core.umath.maximum(
                    self.Mgas * self.qarray[idx] / self.gas_molar_masses[ii], 0.)
                    #conversion to vmr
                self.gas_vmr[self.var_gas_names[ii]] = vmr_tmp
                var_vmr_tot += vmr_tmp
                #print(vmr_tmp, var_vmr_tot, isinstance(vmr_tmp, (np.ndarray)))
            var_vmr_tot = 1. - var_vmr_tot    
            #print(var_vmr_tot)
            for mol, vmr in self.bg_vmr.items():
                #print(mol, vmr, var_vmr_tot, isinstance(var_vmr_tot, (np.ndarray)))
                self.gas_vmr[mol] = vmr * var_vmr_tot

    def update_aerosol_properties(self, atm):
        if self.some_active_aerosol:
            aer_reffs_densities = {}
            #atm.compute_mass_density()
            for ii, idx in enumerate(self.aerosols_idx):
                mmr_rad = np.sqrt(self.qarray[idx,1:] * self.qarray[idx,:-1])
                aer_reffs_densities[self.aerosols_names[ii]] = [self.aerosols_r_eff[ii], 
                    mmr_to_number_density_ratio(mmr_rad, atm.Mgas_rad, self.aerosols_r_eff[ii], self.aerosols_bulk_density[ii])]
        else:
            aer_reffs_densities = None
        return aer_reffs_densities

    def turbulent_diffusion(self, timestep, Htot, atm, cp):
        """Mixes tracers following a diffusion equation
        with a constant Kzz parameter (self.Kzz in m^2/s).

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
                (needs to be converted before it is sent to `turbulent diffusion`)
            Htot: array, np.ndarray
                Total heating rate (in W/kg) of all physical processes
                already computed
            atm: :class:`Atm` object
                The Atm object used in the radiative transfer which
                contains many state variables. 
        """
        new_t = atm.tlay + timestep * Htot
        if self.settings['moist_inhibition']:
            Mgas_tmp = self.Mgas
        else:
            Mgas_tmp = np.mean(self.Mgas)
            Mgas_tmp = np.full_like(self.Mgas, Mgas_tmp)
        qarray = turbulent_diffusion_numba(timestep*cp, self.Nlay,
                    atm.play, atm.plev,
                    atm.dmass, new_t/Mgas_tmp,
                    atm.grav, self.Kzz, self.qarray)
        #self.dm_trac = (qarray - self.qarray) * atm.dmass / (timestep*cp)
        self.qarray = qarray


    def dry_convective_adjustment(self, timestep, Htot, atm, verbose=False):
        """Computes convective adjustement. 

        Parameters
        ----------
            timestep: float
                physical timestep of the current step (in s/cp).
                (needs to be converted before it is sent to `turbulent diffusion)
            Htot: array, np.ndarray
                Total heating rate (in W/kg) of all physical processes
                already computed
            atm: :class:`Atm` object
                The Atm object used in the radiative transfer which
                contains many state variables. 
        """
        new_t = atm.tlay + timestep * Htot
        if self.settings['moist_inhibition']:
            Mgas_tmp = self.Mgas
        else:
            Mgas_tmp = np.mean(self.Mgas)
            Mgas_tmp = np.full_like(self.Mgas, Mgas_tmp)
        H_conv, q_array = dry_convective_adjustment_numba(timestep, self.Nlay, new_t,
                    atm.exner, atm.dmass, self.qarray, Mgas_tmp, verbose=verbose)
        if self.settings['convective_transport']:
            self.qarray=q_array
        return H_conv

    def update_surface_reservoir(self, condensing_pairs_idx = False,
            surf_layer_mass = 0.):
        """Update surface reservoirs of tracers.

        E.g. deals with removing excess condensates from the first layer or
        putting more when everything as been evaporated. 
        """
        if condensing_pairs_idx: # an empty list is False
            qcond_surf_layer = self.settings['qcond_surf_layer']
            for idx_vap, idx_cond in condensing_pairs_idx:
                dm_surf_layer = (qcond_surf_layer - self.qarray[idx_cond, -1]) * surf_layer_mass
                if dm_surf_layer > self.qsurf[idx_cond]: #empties surface reservoir
                    self.qarray[idx_cond, -1] += self.qsurf[idx_cond] / surf_layer_mass
                    self.qsurf[idx_cond] = 0.
                else:  # general case
                    self.qarray[idx_cond, -1] = qcond_surf_layer
                    self.qsurf[idx_cond] -= dm_surf_layer

    def __getitem__(self, key):
        return self.qarray[self.idx[key]]



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
__init__ module to load the exo_k library
"""
from .__version__ import __version__

from exo_k.util.user_func import *
from exo_k.two_stream import two_stream_crisp, two_stream_toon, two_stream_lmdz
from exo_k.util.radiation import *
from exo_k.util.interp import gauss_legendre, split_gauss_legendre
from exo_k.util.cst import *
from exo_k.util.spectrum import Spectrum
from exo_k.util.filenames import *
from exo_k.util.molar_mass import Molar_mass
from .aerosols import Aerosols
from .ktable import Ktable
from .ktable5d import Ktable5d
from .kdatabase import Kdatabase
from .xtable import Xtable
from .atable import Atable, combine_tables
from .adatabase import Adatabase
from .cia_table import Cia_table
from .ciadatabase import CIAdatabase
from .hires_spectrum import Hires_spectrum
from .atm_profile import Atm_profile
from .atm import Atm
from .atm_2band import Atm_2band
from .chemistry import EquChemTable, InterpolationChemistry
from .gas_mix import Gas_mix, Known_composite_species
from .settings import Settings
from .rayleigh import Rayleigh
from exo_k.atm_evolution.atm_evol import Atm_evolution
from exo_k.atm_evolution.condensation import Condensation_Thermodynamical_Parameters, Condensing_species
from exo_k.aerosol.util_aerosol import mmr_to_number_density, mmr_to_number_density_ratio
from .fit import Fit




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
import numba
import exo_k.util.cst as cst
from .condensation import moist_adiabat, compute_condensation_parameters, Tsat_P

@numba.jit(nopython=True, fastmath=True, cache=True)
def dry_convective_adjustment_numba(timestep, Nlay, t_lay, exner, dmass,
    tracer_array, Mgas, verbose = False):
    r"""Computes the heating rates needed to adjust unstable regions 
    of a given atmosphere to a convectively neutral T profile on
    a given timestep.
    
    Parameters
    ----------
        timestep: float
            Duration of the adjustment.
            If given in seconds, H=dT/timestep is in K/s.
            In the current model, timestep is in second over cp.
            This ensures that H=dT/timestep is in W/kg.
        Nlay: int
            Number of atmospheric layers
        t_lay: array, np.ndarray
            Temperatures of the atmospheric layers (K)
        exner: array, np.ndarray
            Exner function computed at the layer centers ((p/psurf)**rcp)

            .. math::
              \Pi=(p / p_{s})^{R/c_p}

        dmass: array, np.ndarray
            Mass of gas in each layer (kg/m^2)

    Returns
    -------
        array
            Heating rate in each atmospheric layer (W/kg).  
    """
    theta_lev=t_lay/exner
    new_theta_lev = np.copy(theta_lev)
    new_Mgas = np.copy(Mgas)
    new_tracers = tracer_array.copy()
    exner_dmass=dmass*exner
    H_conv=np.zeros(Nlay)
    n_iter=0
    if verbose: print('enter convection')
#    if verbose: print(new_theta_lev, Mgas, theta_ov_mu)
    while True:
        theta_ov_mu = new_theta_lev/new_Mgas
        #conv=np.nonzero(new_theta_lev[:-1]-new_theta_lev[1:]<epsilon)[0]
        conv=np.nonzero(theta_ov_mu[:-1]<theta_ov_mu[1:])[0]
        # find convective layers
        if verbose: print(conv)
        #if verbose:
        #    print('start at, end at:',conv[0]-4,conv[-1]+3)
        #    print(theta_ov_mu[conv[0]-4:conv[-1]+4])
        N_conv=conv.size
        if N_conv==0: # no more convective regions, normal exit
            if verbose: print(conv)
            return H_conv, new_tracers
        i_conv=0
        i_top=conv[i_conv] #upper unstable layer
        while i_conv<N_conv-1: #search from the top of the 1st unstable layer for its bottom
            if conv[i_conv+1]==conv[i_conv]+1:
                i_conv+=1
                continue
            else:
                break
        i_bot=conv[i_conv]+1
        if verbose: print('it,ib,:',i_top,i_bot,i_conv)
        mass_conv=0.
        intexner=0.
        theta_mean=0.
        Mmean=0.
        for ii in range(i_top,i_bot+1): # compute new mean potential temperature
            intexner += exner_dmass[ii]
            mass_conv += dmass[ii]
            theta_mean += exner_dmass[ii] * (new_theta_lev[ii] - theta_mean) / intexner
            Mmean += dmass[ii] * (new_Mgas[ii] - Mmean) / mass_conv
        #if verbose: print('theta_mean,Mmean,:',theta_mean,Mmean,theta_mean/Mmean)
        i_top_last, ibot_last = -1, -1
        while (i_top != i_top_last) or (i_bot != ibot_last):
            i_top_last, ibot_last = i_top, i_bot
            while i_top>0:#look for newly unstable layers above
                if theta_ov_mu[i_top-1]<theta_mean/Mmean:
                    i_top -= 1
                    intexner += exner_dmass[i_top]
                    mass_conv += dmass[i_top]
                    theta_mean += exner_dmass[i_top] * (new_theta_lev[i_top] - theta_mean) / intexner
                    Mmean += dmass[i_top] * (new_Mgas[i_top] - Mmean) / mass_conv
                else:
                    break
            while i_bot<Nlay-1: #look for newly unstable layers below
                if theta_ov_mu[i_bot+1]>theta_mean/Mmean:
                    i_bot+=1
                    intexner+=exner_dmass[i_bot]
                    mass_conv+=dmass[i_bot]
                    theta_mean+=exner_dmass[i_bot] * (new_theta_lev[i_bot] - theta_mean) / intexner
                    Mmean += dmass[i_bot] * (new_Mgas[i_bot] - Mmean) / mass_conv
                else:
                    break
        #if verbose: print('it,ib, mconv1,2, M, th:',
        #    i_top,i_bot, mass_conv, np.sum(dmass[i_top:i_bot+1]), Mmean, theta_mean, theta_mean/Mmean)
        # compute heating and adjust before looking for a new potential unstable layer
        H_conv[i_top:i_bot+1] += (theta_mean-new_theta_lev[i_top:i_bot+1]) \
            *exner[i_top:i_bot+1]/timestep
        new_theta_lev[i_top:i_bot+1] = theta_mean
        new_Mgas[i_top:i_bot+1] = Mmean
        # mix tracers
        for q in new_tracers:
            q[i_top:i_bot+1]=np.sum(q[i_top:i_bot+1]*dmass[i_top:i_bot+1])/mass_conv
        n_iter+=1
        if n_iter>Nlay+1:
            if verbose : print('oops, went crazy in convadj')
            break
    return H_conv, new_tracers # we exit through here only when we exceed the max number of iteration

@numba.jit(nopython=True, fastmath=True, cache=True)
def convective_acceleration_numba(timestep, Nlay, H_rad, rad_layers, tau_rad, tau_rads, dmass,
    convective_acceleration_mode = 0, verbose = False):
    r"""Computes a heating rate for the whole convective region to accelerate convergence
    
    Parameters
    ----------
        timestep: float
            Duration of the adjustment.
            If given in seconds, H=dT/timestep is in K/s.
            In the current model, timestep is in second over cp.
            This ensures that H=dT/timestep is in W/kg.
        Nlay: int
            Number of atmospheric layers
        H_rad: array, np.ndarray
            Radiative heating rate
        rad_layers: array, np.ndarray of bool
            Elements in the array are true if layer is purely radiative
        tau_rad: array, np.ndarray
            Baseline radiative timescale for the atmosphere. (e.g. the min of tau_rads)
            Should use the same units as timestep.
        tau_rads: array, np.ndarray
            Radiative timescale for each layer. Should use the same units as timestep.
        dmass: array, np.ndarray
            Mass of gas in each layer (kg/m^2)
        convective_acceleration_mode: int
            convective_acceleration_mode=0 is the default mode
            =1 emulates an old (and bad) behavior.

    Returns
    -------
        array
            Heating rate in each atmospheric layer (W/kg).  
    """
    H_acc = np.zeros(Nlay)
    n_iter=0
    # find convective layers
    conv=np.nonzero(np.invert(rad_layers))[0]
    if verbose: print('in conv acc:',conv)
    N_conv=conv.size
    i_conv=-1
    while True:
        #if verbose:
        #    print('start at, end at:',conv[0]-4,conv[-1]+3)
        #    print(theta_ov_mu[conv[0]-4:conv[-1]+4])
        i_conv += 1
        if i_conv == N_conv: # no more convective regions, normal exit
            if verbose: print('normal exit')
            return H_acc
        i_top=conv[i_conv] #upper unstable layer
        while i_conv<N_conv-1: #search from the top of the 1st unstable layer for its bottom
            if conv[i_conv+1]==conv[i_conv]+1:
                i_conv+=1
                continue
            else:
                break
        i_bot=conv[i_conv]
        if verbose: print('it,ib,:',i_top,i_bot,i_conv)
        mass_conv=0.
        H_rad_mean=0.
        for ii in range(i_top,i_bot+1): # computeaverage radiative heating
            mass_conv += dmass[ii]
            H_rad_mean += dmass[ii] * (H_rad[ii] - H_rad_mean) / mass_conv
        tau = np.amin(tau_rads[i_top:i_bot+1])
        # compute heating and adjust before looking for a new potential unstable layer
        if convective_acceleration_mode == 0:
            H_acc[i_top:i_bot+1] += H_rad_mean * tau / tau_rad
        else:
            H_acc[i_top:i_bot+1] += H_rad_mean * tau / timestep
        n_iter+=1
        if n_iter>Nlay+1:
            if verbose : print('oops, went crazy in convective_acceleration')
            break
    return H_acc # we exit through here only when we exceed the max number of iteration

@numba.jit(nopython=True, fastmath=True, cache=True)
def turbulent_diffusion_numba(timestep, Nlay, p_lay, p_lev, dmass,
        t_lay_ov_mu, g, Kzz, tracer_array, verbose = False):
    r"""Solves turbulent diffusion equation:

    .. math::
      \rho frac{\partial q}{\partial t} = \frac{\partial F_{diff}}{\partial z}
    
    with a diffusive flux given by 

    .. math::
      F_{diff} = - \rho K_{zz} \frac{\partial q}{\partial z}

    The equation is solved with an implicit scheme assuming that
    there is no flux at the top and bottom boundaries
    (evaporation must be treated separately for now).

    Parameters
    ----------
        timestep: float
            Time step in seconds.
        Nlay: int
            Number of atmospheric layers
        t_lay_ov_mu: array, np.ndarray
            Temperatures of the atmospheric layers divided by the molar_mass in kg/mol
        p_lay: array, np.ndarray
            Pressure at the layer centers (Pa)
        p_lev: array, np.ndarray
            Presure at the Nlay+1 level boundaries (Pa)
        dmass: array, np.ndarray
            Mass of gas in each layer (kg/m^2)
        g: float
            Gravity (m/s^2)
        Kzz: float
            Eddy mixing coefficient (m^2/s)
        tracer_array: array, np.ndarray (Ntrac, Nlay)
            Array containing the mass mixing ratio of all tracers at each layer
            before the mixing

    Returns
    -------
        new_tracers: array, np.ndarray (Ntrac, Nlay)
            Array containing the mass mixing ratio of all tracers at each layer
            after the mixing
    """
    mid_density = p_lev[1:-1]*2./(cst.RGP*(t_lay_ov_mu[1:]+t_lay_ov_mu[:-1]))
    mid_factor = - g * g * timestep * mid_density**2 / np.diff(p_lay) * 0.5*(Kzz[1:]+Kzz[:-1])
    if verbose:
        print(mid_factor)
        print(dmass)
    A = np.zeros(Nlay)
    B = np.copy(dmass)
    C = np.zeros(Nlay)
    A[1:] = mid_factor
    C[:-1] = mid_factor
    B += - C - A
    D = np.zeros(Nlay)
    new_tracers = tracer_array.copy()
    for i_q, q in enumerate(new_tracers):
        D = dmass * q
        new_q = DTRIDGL(Nlay,A,B,C,D)
        new_tracers[i_q] = new_q
        #mix_rate[name] = (new_q-q)/timestep
    return new_tracers


@numba.jit(nopython=True, fastmath=True, cache=True)
def compute_condensation_numba(timestep, Nlay, tlay, play, cp, Mgas, qarray,
        idx_vap, idx_cond, thermo_parameters, latent_heating = True,
        condensation_timestep_reducer = None,
        verbose = False):
    r"""Computes the heating rates needed to bring sursaturated 
    regions back to saturation

    Parameters
    ----------
        timestep
        Nlay: float
            Number of layers
        tlay: array, np.ndarray
            Layer temperatures
        cp: float
            Specific heat capacity
        Mgas: array, np.ndarray
            Layer molar mass
        qarray: array, np.ndarray
            Array of tracer specific concentrations
        idx_vap, idx_cond: int
            Indices for condensing species in tracer array
        thermo_parameters: array, np.ndarray
            Array of thermodynamical paramaters for condensing species.
            See `:class:`~exo_k.atm_evolution.condensation.Condensation_Thermodynamical_Parameters`
        latent_heating: bool
            Whether latent heating should be taken into account

    Returns
    -------
        H_cond: array, np.ndarray
            Heating rate in W/kg
    """
    itermax = 1000
    if condensation_timestep_reducer is None:
        alpha = 1.
    else:
        alpha = condensation_timestep_reducer
    H_cond = np.zeros(Nlay)
    if latent_heating:
        for ii in range(Nlay):
            i_iter = 0
            T_tmp = tlay[ii]
            qvap = qarray[idx_vap,ii]
            qcond = qarray[idx_cond,ii]
            if qvap == 0.: continue
            while i_iter < itermax:
                Lvap, qsat, dqsat_dt = compute_condensation_parameters(T_tmp, play[ii], Mgas[ii], 
                        thermo_parameters[1], thermo_parameters[2], thermo_parameters[3], thermo_parameters[4],
                        thermo_parameters[5], thermo_parameters[6], thermo_parameters[7], thermo_parameters[8],
                        thermo_parameters[9])
                dqvap = alpha * (qsat - qvap) / (1. + Lvap * dqsat_dt / cp)
                if dqvap > qcond:
                    dqvap = qcond
                qcond -= dqvap
                qvap += dqvap
                T_tmp -= Lvap * dqvap / cp
                if np.abs(dqvap/(qvap*alpha)) <= 1.e-9: break
                i_iter += 1
            H_cond[ii] = (T_tmp - tlay[ii]) / timestep
            qarray[idx_vap,ii] = qvap
            qarray[idx_cond,ii] = qcond
            if verbose: print('in cond, i, iter:', ii, i_iter+1, dqvap/qvap)
            if verbose: print('in cond, RH, T:', qvap/qsat, T_tmp, tlay[ii], dqsat_dt)

    else:
        # here we treat whole arrays at once.
        Lvap, qsat, dqsat_dt = compute_condensation_parameters(tlay, play, Mgas, 
                thermo_parameters[1], thermo_parameters[2], thermo_parameters[3], thermo_parameters[4],
                thermo_parameters[5], thermo_parameters[6], thermo_parameters[7], thermo_parameters[8],
                thermo_parameters[9])
        qvap = qarray[idx_vap]
        dqvap= np.where(qsat <= qvap, qsat-qvap, 0.)
        qarray[idx_vap] += dqvap
        qarray[idx_cond] -= dqvap
        if verbose: print('in cond, RH, T:', qarray[idx_vap]/qsat, tlay, dqsat_dt)
    return H_cond

@numba.jit(nopython=True, fastmath=True, cache=True)
def moist_convective_adjustment_numba(timestep, Nlay, tlay, play, dmass, cp, Mgas, q_array,
        i_vap, i_cond, thermo_parameters, 
        moist_inhibition = True, verbose = False):
    r"""Computes the heating rates needed to adjust unstable regions 
    of a given atmosphere to a moist adiabat.

    Parameters
    ----------
        timestep
        Nlay: float
            Number of layers
        tlay: array, np.ndarray
            Layer temperatures
        play:array
            Pressure at layer centers
        dmass: array, np.ndarray
            mass of layers in kg/m^2
        cp: float
            specific heat capacity at constant pressure
        q_array: array, np.ndarray
            mass mixing ratio of tracers
        i_vap: int
            index of vapor tracer in qarray
        i_cond: int
            index of condensate tracer in qarray
        qsat: array, np.ndarray
            Saturation mmr for each layer
        dqsat_dt: array, np.ndarray
            d qsat / dT in each layer
        Lvap: array, np.ndarray
            Latent heat of vaporization (can have different values
            in each layer if Lvap=f(T))
        dlnt_dlnp_moist: array, np.ndarray
            threshold thermal gradient (d ln T / d ln P) for a moist atmosphere
            computed at the layer centers.
        q_crit: array, np.ndarray
            Critical mass mixing ratio for the inhibition of moist convection
            (Eq. 17 of Leconte et al. 2017)

    Returns
    -------
        H_madj: array, np.ndarray
            Heating rate in each atmospheric layer (W/kg). 
        new_q: array, np.ndarray
            tracer mmr array after adjustment.
        new_t: array, np.ndarray
            Temperature of layers after adjustment. 
    """
    if verbose: print('enter moist adj')
    H_madj=np.zeros(Nlay)
    new_q = q_array.copy()
    new_t = tlay.copy()
    dp = np.diff(play)

    dlnt_dlnp_moist, Lvap, psat, qsat, dqsat_dt, q_crit = \
        moist_adiabat(new_t, play, cp, Mgas, thermo_parameters[0],
            thermo_parameters[1], thermo_parameters[2], thermo_parameters[3], thermo_parameters[4],
            thermo_parameters[5], thermo_parameters[6], thermo_parameters[7], thermo_parameters[8],
            thermo_parameters[9])

    nabla_interlayer = tlay * dlnt_dlnp_moist /play
    nabla_interlayer = 0.5*(nabla_interlayer[:-1]+nabla_interlayer[1:])
    dTmoist_array = nabla_interlayer * dp
    dT_inter_lay = np.diff(tlay)
    qvap = new_q[i_vap]
    mvap_sursat_array = (qvap-qsat) * dmass
    if moist_inhibition:
        q_crit_criterion = qvap/q_crit < 1. # convection possible if True. qcri can be negative.
    else:
        q_crit_criterion = qvap<2. #should always be true
    #print('dT:', dT_inter_lay)
    #print('dTmoist:', dTmoist_array)
    #dT_unstab = np.nonzero(dT_inter_lay>dTmoist_array)[0]
    #saturated = np.nonzero(mvap_sursat_array>0.)[0]
    conv = np.nonzero((dT_inter_lay>dTmoist_array)*(mvap_sursat_array[:-1]>0.) \
            *q_crit_criterion[:-1])[0]# find convective layers
    if verbose: 
        print(conv)
        print(np.nonzero(dT_inter_lay>dTmoist_array)[0])
        print(np.nonzero(mvap_sursat_array[:-1]>0.)[0])
        print(np.nonzero(q_crit_criterion)[0])
    N_conv=conv.size
    if N_conv==0: # no more convective regions, can exit
        return H_madj, new_q, new_t
    i_top=conv[0] #upper unstable layer
    T_top = new_t[i_top]
    mvap_sursat = mvap_sursat_array[i_top]
    dqsdm = dqsat_dt[i_top]*dmass[i_top]
    int_dqsdm = dqsdm
    C = cp*dmass[i_top] + Lvap[i_top]*dqsdm
    B = C * new_t[i_top] + Lvap[i_top] * mvap_sursat_array[i_top]
    dT_moist = 0.
    int_m_cond = mvap_sursat_array[i_top] + dqsdm*(new_t[i_top] - dT_moist)
    i_bot=i_top+1
    while i_bot<Nlay: #search for the bottom of the 1st unstable layer from its top
        tmp_sursat = mvap_sursat + mvap_sursat_array[i_bot]
        tmp_dT_moist = dT_moist + dTmoist_array[i_bot-1]
        dqsdm = dqsat_dt[i_bot] * dmass[i_bot]
        tmp_int_dqsdm = int_dqsdm + dqsdm
        tmp_int_m_cond = int_m_cond + mvap_sursat_array[i_bot] + dqsdm * (new_t[i_bot] - tmp_dT_moist)
        tmp = cp *dmass[i_bot] + Lvap[i_bot]* dqsdm
        tmp_C = C + tmp
        tmp_B = B + tmp * (new_t[i_bot]-tmp_dT_moist) + Lvap[i_bot] * mvap_sursat_array[i_bot]
        tmp_new_Ttop = tmp_B / tmp_C
        tmp_m_cond = tmp_int_m_cond - tmp_int_dqsdm * tmp_new_Ttop
        if tmp_sursat>0. and tmp_dT_moist<new_t[i_bot]-T_top and q_crit_criterion[i_bot] and tmp_m_cond>0.:
            dT_moist = tmp_dT_moist
            mvap_sursat = tmp_sursat
            int_dqsdm = tmp_int_dqsdm
            int_m_cond = tmp_int_m_cond
            C = tmp_C
            B = tmp_B
            m_cond = tmp_m_cond
            i_bot += 1
            continue
        else:
            break
    i_bot -= 1
    if verbose: print('it,ib=', i_top, i_bot)
    if i_top == i_bot: # need at least 2 layers to convect, so exit
        return H_madj, new_q, new_t
    new_Ttop = B / C
    if verbose: print(new_Ttop, m_cond, dT_moist)
    dT = new_Ttop - new_t[i_top]
    qvap[i_top] = qsat[i_top] + dqsat_dt[i_top] * dT
    #if verbose: print('top i, dT, qv, qs, dqs, qf', i_top, dT, q_array[i_vap, i_top], qsat[i_top], dqsat_dt[i_top], qvap[i_top])
    new_t[i_top] = new_Ttop
    H_madj[i_top] = dT / timestep
    for ii in range(i_top+1, i_bot+1):
        dT = new_t[ii-1] + dTmoist_array[ii-1] - new_t[ii]
        #print(ii, new_t[ii-1], dTmoist_array[ii-1],  new_t[ii], new_t[ii-1] + dTmoist_array[ii-1] - new_t[ii])
        qvap[ii] = qsat[ii] + dqsat_dt[ii] * dT
        new_t[ii] += dT
        # compute heating and adjust before looking for a new potential unstable layer
        H_madj[ii] = dT / timestep
        #if verbose: print('i, dT, qv, qs, dqs, qf', ii, dT, q_array[i_vap, ii], qsat[ii], dqsat_dt[ii], qvap[ii])
    # put ice
    m_cond_2 = np.sum((q_array[i_vap, i_top:i_bot+1]-qvap[i_top:i_bot+1])*dmass[i_top:i_bot+1])
    dTmoist_array[i_top-1]=0.
    if m_cond<0.:
        print('Negative condensates in moist adj, i:', i_top, i_bot)
        print(q_array[i_vap, i_top:i_bot+1], qvap[i_top:i_bot+1], q_array[i_vap, i_top:i_bot+1]-qvap[i_top:i_bot+1])
    m_conv = np.sum(dmass[i_top:i_bot+1])
    new_q[i_cond, i_top:i_bot+1] += m_cond / m_conv
    if verbose: 
        print('m_cond, m_conv, m_cond2', m_cond, m_conv, m_cond_2)
    return H_madj, new_q, new_t

@numba.jit(nopython=True, fastmath=True, cache=True)
def compute_rainout_numba(timestep, Nlay, tlay, play, dmass, cp, Mgas, qarray,
        idx_vap, idx_cond, thermo_parameters, evap_coeff, qvap_deep, q_cloud=0.,
        latent_heating=False,
        verbose = False):
    r"""Computes the heating rates needed to adjust unstable regions 
    of a given atmosphere to a moist adiabat.

    Parameters
    ----------
        timestep
        Nlay: float
            Number of layers
        tlay: array, np.ndarray
            Layer temperatures
    """
    H_rain=np.zeros(Nlay)
    Lvap, qsat, dqsat_dt = compute_condensation_parameters(tlay, play, Mgas, 
            thermo_parameters[1], thermo_parameters[2], thermo_parameters[3], thermo_parameters[4],
            thermo_parameters[5], thermo_parameters[6], thermo_parameters[7], thermo_parameters[8],
            thermo_parameters[9])
    if not latent_heating:
        Lvap = Lvap * 0.
    Tsat_p = Tsat_P(play, thermo_parameters[5], thermo_parameters[8], thermo_parameters[9])

    if verbose: print('in rainout, RH, T, qice:', qarray[idx_vap]/qsat,  tlay, qarray[idx_cond])
    mass_cond = 0.
    for i_lay in range(Nlay):
        #  if evap_coeff =1, rain vaporisation in an undersaturated layer can fill the layer up to the (estimated) saturation
        #  if 0 < evap_coeff < 1, rain vaporisation in one layer is limited to a fraction of the amount that would saturate the layer
        #  This allows not to exceed saturation, to spread rain vaporization in more and denser layers
        qvap = qarray[idx_vap,i_lay]
        if (tlay[i_lay] >= Tsat_p[i_lay]): # above boiling temperature, try to evaporate everything
            dqvap = (qsat[i_lay] - qvap) / (1. + Lvap[i_lay]*dqsat_dt[i_lay]/cp)
        else:
            dqvap = evap_coeff * (qsat[i_lay] - qvap) / (1. + Lvap[i_lay]*dqsat_dt[i_lay]/cp)
        dqvap = np.core.umath.minimum(dqvap, 1.- qvap)
        if qarray[idx_cond,i_lay]>=q_cloud:
            mass_cond += (qarray[idx_cond,i_lay] - q_cloud) * dmass[i_lay]
            qarray[idx_cond,i_lay] = q_cloud
        # dqvap is the change in vapor content to reach saturation and accounting for the temperature change.
        # dqvap < 0 implies condensation, meaning that there is a remaining excess of vapor after the previous 
        # condensation step. In such case we apply this new change in vapor content and temperature and
        # increase the amount of falling condensed species (mass_cond).
        # dqvap > 0 implies evaporation. Here there are two possibilities:
        #   - the amount of condensates is lower than dqvap. All condensates are vaporised in the layer
        #    and the tempertaure change is -Ldqv/cp where dqv is the actual change in vapor.
        #   - the amount of condensates is larger than dqvap. We then apply dqvap and the corresponding 
        #    change in temperature and transfer the remaining condensate in the falling rain reservoir.
        if dqvap < 0: # more condensation in the layer
            qarray[idx_vap][i_lay] += dqvap
            H_rain[i_lay] = -Lvap[i_lay]*dqvap/(cp*timestep)
            mass_cond -= dqvap*dmass[i_lay]               
            if verbose: print('dqvap < 0:',i_lay, dqvap, qvap, mass_cond, mass_cond/dmass[i_lay])
        else: # evaporation of rain
            mass_dvap = dqvap*dmass[i_lay]     
            if mass_dvap > mass_cond: # evaporate everything
                if verbose: print('mass_dvap > mass_cond:', i_lay, dqvap, qvap, mass_cond, mass_cond/dmass[i_lay], mass_dvap, dmass[i_lay],(mass_dvap > mass_cond), (tlay[i_lay] >= Tsat_p[i_lay]))
                qarray[idx_vap,i_lay] += mass_cond/dmass[i_lay]
                H_rain[i_lay] = - Lvap[i_lay] * mass_cond / (dmass[i_lay]*cp*timestep)
                mass_cond = 0.
            else:
                if verbose: print('mass_dvap < mass_cond:', i_lay, dqvap, qvap, mass_cond, mass_cond/dmass[i_lay])
                qarray[idx_vap,i_lay] += dqvap
                H_rain[i_lay] = -Lvap[i_lay]*dqvap/(cp*timestep)
                mass_cond -= dqvap*dmass[i_lay]
    if mass_cond != 0.:
        if verbose: print('mass_cond=',mass_cond)
        qarray[idx_cond,-1]+= mass_cond/dmass[-1]
        # Issue : qarray[idx_cond,-1] sometimes can become large (when starting with a hot atmosphere dominated with H2O that cools to saturation)
    if qvap_deep>=0.:
        qarray[idx_vap,-1] = qvap_deep
    return H_rain



@numba.jit(nopython=True, fastmath=True, cache=True)
def molecular_diffusion_numba(timestep, Nlay, p_lay, p_lev, dmass,
        tlay, mu, g, Dmol, verbose = False):
    r"""Solves turbulent diffusion equation:

    .. math::
      \rho frac{\partialT}{\partial t} = \frac{\partial F_{diff}}{\partial z}
    
    with a diffusive flux given by 

    .. math::
      F_{diff} = - \rho D_{mol} \frac{\partial T}{\partial z}

    The equation is solved with an implicit scheme assuming that
    there is no flux at the top and bottom boundaries
    (evaporation must be treated separately for now).

    Parameters
    ----------
        timestep: float
            Time step in seconds.
        Nlay: int
            Number of atmospheric layers
        t_lay_ov_mu: array, np.ndarray
            Temperatures of the atmospheric layers divided by the molar_mass in kg/mol
        p_lay: array, np.ndarray
            Pressure at the layer centers (Pa)
        p_lev: array, np.ndarray
            Presure at the Nlay+1 level boundaries (Pa)
        dmass: array, np.ndarray
            Mass of gas in each layer (kg/m^2)
        g: float
            Gravity (m/s^2)
        Dmol: float
            molecular diffusion coefficient (m^2/s)

    Returns
    -------
        new_tlay: array, np.ndarray (Nlay)
            Array containing the temperature at each layer
            after the mixing
    """
    mid_density = p_lev[1:-1]*2.*(mu[1:]+mu[:-1])/(cst.RGP*(tlay[1:]+tlay[:-1]))
    mid_factor = - g * g * timestep * mid_density**2 / np.diff(p_lay) * Dmol
    if verbose:
        print(mid_factor)
        print(dmass)
    A = np.zeros(Nlay)
    B = np.copy(dmass)
    C = np.zeros(Nlay)
    A[1:] = mid_factor
    C[:-1] = mid_factor
    B += - C - A
    D = dmass * tlay
    new_tlay = DTRIDGL(Nlay,A,B,C,D)
    H_diff = (new_tlay-tlay)/timestep
    return H_diff

@numba.jit(nopython=True, fastmath=True, cache=True)
def DTRIDGL(L,AF,BF,CF,DF):
    """
    !  GCM2.0  Feb 2003

    !     DOUBLE PRECISION VERSION OF TRIDGL

          DIMENSION AF(L),BF(L),CF(L),DF(L),XK(L)
          DIMENSION AS(2*L),DS(2*L)

    !*    THIS SUBROUTINE SOLVES A SYSTEM OF TRIDIAGIONAL MATRIX
    !*    EQUATIONS. THE FORM OF THE EQUATIONS ARE:
    !*    A(I)*X(I-1) + B(I)*X(I) + C(I)*X(I+1) = D(I)

    !======================================================================!
    """
    AS=np.empty_like(AF)
    DS=np.empty_like(AF)
    XK=np.empty_like(AF)
    AS[-1] = AF[-1]/BF[-1]
    DS[-1] = DF[-1]/BF[-1]

    for I in range(1,L):
        X         = 1./(BF[L+1-I-2] - CF[L+1-I-2]*AS[L+2-I-2])
        AS[L+1-I-2] = AF[L+1-I-2]*X
        DS[L+1-I-2] = (DF[L+1-I-2]-CF[L+1-I-2]*DS[L+2-I-2])*X
 
    XK[0]=DS[0]
    for I in range(1,L):
        XKB   = XK[I-1]
        XK[I] = DS[I]-AS[I]*XKB
    return XK
    


# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

This module contain classes to fit a spectrum
"""
import copy
import time
import numpy as np
import scipy.optimize as sciopt
from .atm import Atm
from .util.cst import G


class Fit(object):
    """Class to fit spectra
    """

    def __init__(self, spectrum_to_fit=None, uncertainties=None, fitting_parameters=None, **kwargs):
        """Initialization method. 
        """
        self.cost_function = None
        self.atm = None
        self.N_iter = 0
        self.spectrum_to_fit = None
        self.set_spectrum_to_fit(spectrum_to_fit=spectrum_to_fit, uncertainties=uncertainties)
        
        if fitting_parameters is not None:
            self.set_fitting_parameters(fitting_parameters=fitting_parameters)


    def init_atmosphere(self, composition=None, Mp=None, **kwargs):
        """initializes atmospheric model
 
        Parameters
        ----------
            Mp: float
                Mass of the planet in Kg.
        """
        if composition is None:
            raise RuntimeError('composition should not be None')
        self.atm = Atm(composition=composition, **kwargs)
        self.composition = copy.deepcopy(composition)
        self.Mp = Mp

    def set_fitting_parameters(self, fitting_parameters=None):
        """
        Parameters
        ----------
            fitting_paramaters: list of str
                The parameters to fit.
                Must be in T, Rp, logx_MOL
        """
        if fitting_parameters is None:
            raise RuntimeError('fitting_parameters should not be None.')
        self.fitting_parameters = fitting_parameters
        self.Nparam = len(self.fitting_parameters)

    def set_spectrum_to_fit(self, spectrum_to_fit=None, uncertainties=None):
        """
        Parameters
        ----------
            spectrum_to_fit: :class:`Spectrum` object
                The spectrum to fit
            uncertainties: float or array
                The uncertainty on each spectral point. 
        """
        if spectrum_to_fit is None:
            raise RuntimeError('spectrum_to_fit should not be None.')
        self.spectrum_to_fit = spectrum_to_fit
        if uncertainties is None:
            raise RuntimeError('uncertainties should not be None.')
        self.uncertainties = uncertainties
        self.Nw = self.spectrum_to_fit.value.size
    
    def plot_spectrum_to_fit(self, ax, capsize=0, fmt='k.', ecolor='.7',
            ylabel='Depth', xlabel='Wavelength', **kwargs):
        """Plots the spectrum that is fitted with uncertainties
        """
        ax.errorbar(self.spectrum_to_fit.wls, self.spectrum_to_fit.value, yerr=self.uncertainties,
                capsize=capsize, fmt=fmt, ecolor=ecolor)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        

    def transmission_spectrum(self, parameters, rebin_spectrum=False, **kwargs):
        """Method to compute the spectrum with a new set of parameters
        """
        var_species = 0
        var_comp = self.composition
        for ii, param in enumerate(self.fitting_parameters):
            if param == 'T':
                tlay = np.ones(self.atm.Nlay) * parameters[ii]
                self.atm.set_T_profile(tlay)
            if param.startswith('logx_'):
                mol = param[5:]
                var_comp[mol] = np.power(10., parameters[ii])
                var_species += 1
            if param == 'Rp':
                self.atm.set_Rp(parameters[ii])
                if self.Mp is not None:
                    grav = G * self.Mp / parameters[ii]**2
                    self.atm.set_grav(grav=grav)
        if var_species > 0:
            self.atm.set_gas(var_comp)
        spectrum = self.atm.transmission_spectrum(**kwargs)
        if rebin_spectrum: spectrum.bin_down(self.spectrum_to_fit.wnedges)
        return spectrum

    def set_cost_function(self):
        """
        Defines the cost (or merit) function to minimize. 
        """
        if self.spectrum_to_fit is None:
            raise RuntimeError('Use set_spectrum_to_fit before setting cost function.')
        if self.atm is None:
            raise RuntimeError('Use init_atmosphere before setting cost function.')
        if self.atm.k_database is None:
            raise RuntimeError('A k_database should be provided to init_atmosphere.')
        self.rebin_spectrum = np.any(self.atm.k_database.wnedges != self.spectrum_to_fit.wnedges)
        def distance(parameters):
            self.last_spec = self.transmission_spectrum(parameters, rebin_spectrum=self.rebin_spectrum)
            return np.sum(((self.last_spec.value - self.spectrum_to_fit.value)/self.uncertainties)**2)/self.Nw
        self.cost_function = distance

    def set_progress_report(self, Nprint=None):
        if Nprint is not None:
            self.Nprint = Nprint
            self.N_iter = 0
            def callback(parameters):
                if self.N_iter % self.Nprint == 0:
                    res=dict()
                    for ii, param in enumerate(self.fitting_parameters):
                        res[param] = parameters[ii]
                    print('Niter=',self.N_iter, res)
                self.N_iter += 1
            self.callback = callback
        else:
            self.callback = None
        return None
    
    def minimize(self, initial_guess=None, bounds=None, method='Nelder-Mead',
            verbose=False, tol=.1e0, Nprint=None, **kwargs):
        """
        Minimizes the cost function over the fitting parameters. 

        Parameters
        ----------
            initial_guess: list
                Initial value for the fitting parameters.
                Must be in the order declared in fitting_parameters.
            bounds: list or 2-value arrays
                Lower and upper values for the parameters.
                Must be in the order declared in fitting_parameters.
            method: str
                Method used for minimization. 
                For now, only 'Nelder-Mead' seems to handle bounds and yield relatively
                good results.
            tol: float
                tolerance transmitted to sciopt.minimize
        """        
        self.set_bounds(bounds=bounds)
        if initial_guess is None:
            initial_guess = np.zeros(self.Nparam)
            for ii in range(self.Nparam):
                initial_guess[ii] = 0.5 * (self.bounds[ii][0] + self.bounds[ii][1])
        self.set_cost_function()
        start = time.time()
        if verbose:
            print(self.cost_function, initial_guess,
            method, bounds,
            tol)
        self.set_progress_report(Nprint)
        self.full_minimize_output = sciopt.minimize(self.cost_function, initial_guess,
            method=method, bounds=self.bounds,
            tol=tol, callback=self.callback, **kwargs)
        self.best_fit_parameters = self.full_minimize_output.x
        if verbose:
            print('computation time: ', time.time()-start,'s')
            print(self.best_fit())
            print(self.full_minimize_output)
        return self.best_fit()

    def best_fit(self):
        """Output last fit results in a dictionary
        """
        res = dict()
        for ii, param in enumerate(self.fitting_parameters):
            res[param] = self.best_fit_parameters[ii]
        return res

    def best_fit_spectrum(self, rebin_spectrum=False):
        """Returns the spectrum computed with the best fit parameters
        """
        spectrum = self.transmission_spectrum(self.best_fit_parameters, rebin_spectrum=rebin_spectrum)
        return spectrum

    def contributions(self, rebin_spectrum=False):
        """Computes a spectrum with each molecule contribution
        """
        molecules = set(self.atm.gas_mix.composition.keys())
        molecules = molecules.intersection(self.atm.gas_mix.k_database.molecules) 
        spectra = dict()
        for mol in molecules:
            inactive_mols = list(self.atm.gas_mix.composition.keys()-{mol})
            spectra[mol] = self.transmission_spectrum(self.best_fit_parameters, rebin_spectrum=rebin_spectrum,
                    inactive_molecules=inactive_mols)
        return spectra

    def set_log_likelihood(self):
        """
        Defines the log likelihood. 
        """
        if self.spectrum_to_fit is None:
            raise RuntimeError('Use set_spectrum_to_fit before setting cost function.')
        if self.atm is None:
            raise RuntimeError('Use init_atmosphere before setting cost function.')
        if self.atm.k_database is None:
            raise RuntimeError('A k_database should be provided to init_atmosphere.')
        self.rebin_spectrum = np.any(self.atm.k_database.wnedges != self.spectrum_to_fit.wnedges)
        def loglike(parameters):
            self.last_spec = self.transmission_spectrum(parameters, rebin_spectrum=self.rebin_spectrum)
            return -0.5 * np.sum(((self.last_spec.value - self.spectrum_to_fit.value)/self.uncertainties)**2)
        self.loglike = loglike
    
    def set_bounds(self, bounds=None):
        if bounds is None:
            print("""You should provide bounds in the form of a dictionary
where keys are the fitting parameters and values are
2 value arrays with the lower and upper bounds.""")
            raise RuntimeError('bounds should not be None.')
        self.upper_bounds = np.ones(self.Nparam)
        self.lower_bounds = np.zeros(self.Nparam)
        self.bounds = list()
        for ii, param in enumerate(self.fitting_parameters):
            try:
                self.upper_bounds[ii] = bounds[param][1]
                self.lower_bounds[ii] = bounds[param][0]
                self.bounds.append([bounds[param][0], bounds[param][1]])
            except:
                self.upper_bounds[ii] = bounds['logx'][1]
                self.lower_bounds[ii] = bounds['logx'][0]
                self.bounds.append([bounds['logx'][0], bounds['logx'][1]])

    def set_prior_transform(self, bounds=None):
        """
        Defines the piors. 
        """
        self.set_bounds(bounds=bounds)
        self.delta_bounds = self.upper_bounds-self.lower_bounds

        def prior_transform(x):
            return self.delta_bounds * x + self.lower_bounds
        self.prior_transform = prior_transform

    def sample(self, bounds=None, method='single',
            npoints=10, maxcall=None,
            verbose=False, **kwargs):
        """
        Minimizes the cost function over the fitting parameters. 

        Parameters
        ----------
            initial_guess: list
                Initial value for the fitting parameters.
                Must be in the order declared in fitting_parameters.
            bounds: list or 2-value arrays
                Lower and upper values for the parameters.
                Must be in the order declared in fitting_parameters.
            method: str
                Method used for minimization. 
                For now, only 'Nelder-Mead' seems to handle bounds and yield relatively
                good results.
            tol: float
                tolerance transmitted to sciopt.minimize
        """        
        import nestle
        self.set_log_likelihood()
        self.set_prior_transform(bounds=bounds)
        start = time.time()
        self.full_sample_output = nestle.sample(self.loglike, self.prior_transform,
                self.Nparam, method=method, npoints=npoints, maxcall=maxcall, **kwargs)

        # weighted average and covariance:
        self.best_fit_parameters, self.covariance = nestle.mean_and_cov( \
                    self.full_sample_output.samples, self.full_sample_output.weights)

        if verbose:
            print('computation time: ', time.time()-start,'s')
            #print(self.full_sample_output.summary)
            for ii, param in enumerate(self.fitting_parameters):
                print(param+" = {0:5.2f} +/- {1:5.2f}".format( \
                    self.best_fit_parameters[ii], np.sqrt(self.covariance[ii, ii])))

        return self.best_fit()



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
import numba
from exo_k.util.cst import PI, N_A


@numba.jit(nopython=True, fastmath=True, cache=True)
def mmr_to_number_density(mmr, gas_density, r_eff, aerosols_bulk_density):
    """Converts a mass mixing ratio (mmr or q) in a number density of particles
    (in number per unit volume)

    Parameters
    ----------
        mmr: float or array
            Mass mixing ratio (in kg per kg of air)
        gas_density: float or array
            Density of the gas (in kg/m^3)
        r_eff: float or array
            Effective radius of the particles (m)
        aerosols_bulk_density: float or array
            Density of the constituent of the condensed particles (in kg/m^3)
    """
    particle_mass = 4. * PI * r_eff**3 * aerosols_bulk_density / 3.
    #print(particle_mass.shape, mmr.shape, gas_density.shape)
    return mmr * gas_density / particle_mass


@numba.jit(nopython=True, fastmath=True, cache=True)
def mmr_to_number_density_ratio(mmr, Mgas, r_eff, aerosols_bulk_density):
    """Converts a mass mixing ratio (mmr or q) in a ratio between particles density
    and molecules density

    Parameters
    ----------
        mmr: float or array
            Mass mixing ratio (in kg per kg of air)
        Mgas: float or array
            gas molar mass (in kg/mol)
        r_eff: float or array
            Effective radius of the particles (m)
        aerosols_bulk_density: float or array
            Density of the constituent of the condensed particles (in kg/m^3)
    """
    particle_mass = 4. * PI * r_eff**3 * aerosols_bulk_density / 3.
    #print(particle_mass.shape, mmr.shape, gas_density.shape)
    return mmr * Mgas / (particle_mass * N_A)



"""
@author: jeremy leconte

A class with some basic functions for all objects with a spectral dimension
"""
import numpy as np
from .cst import PI
from .radiation import Bnu_integral_num, Bnu

class Spectral_object(object):
    """A class with some basic functions for all objects with a spectral dimension
    """

    def __init__(self):
        """Initializes some attributes to None
        """
        self.wns = None
        self.wnedges = None
        self.Nw = None
        self.wn_unit='cm^-1'

    @property
    def wls(self):
        """Returns the wavelength array for the bin centers (in micron)
        """
        if self.wns is not None:
            return 10000./self.wns
        else:
            raise RuntimeError('self.wns should not be None when requesting wls')

    @property
    def wledges(self):
        """Returns the wavelength array for the bin edges (in micron)
        """
        if self.wnedges is not None:
            return 10000./self.wnedges
        else:
            raise RuntimeError('self.wnedges should not be None when requesting wledges')

    @property
    def wnrange(self):
        """Returns the limits of the wavenumber range.

        First tries with wnedges (for Ktables) and then wns (Xtables).
        """
        if self.wnedges is not None:
            return self.wnedges[[0,-1]]
        elif self.wns is not None:
            return self.wns[[0,-1]]
        else:
            raise RuntimeError('self.wns or wnedges should not be None when requesting wnrange')

    @property
    def wlrange(self):
        """Returns the limits of the wavelength range
        """
        if (self.wnedges is None) and (self.wns is None):
            raise RuntimeError('self.wns or wnedges should not be None when requesting wlrange')
        else:
            return np.sort(10000./self.wnrange)

    def _compute_spectral_range(self, wn_range=None, wl_range=None):
        """Converts an unordered spectral range in either wavenumber or wavelength
        in an ordered wavenumber range.

        Parameters
        ----------
            wn_range: list or array of size 2
                Minimum and maximum wavenumber (in cm^-1).
            wl_range: list or array of size 2
                Minimum and maximum wavelength (in micron)
        """
        if wl_range is not None:
            if wn_range is not None:
                print('Cannot specify both wl and wn range!')
                raise RuntimeError()
            else:
                _wn_range=np.sort(10000./np.array(wl_range))
        else:
            if wn_range is not None:
                _wn_range=np.sort(np.array(wn_range))
            else:
                _wn_range=self._wn_range
        return _wn_range

    def select_spectral_range(self, wn_range=None, wl_range=None):
        """Select spectral range, without restricting the data
        (but the spectral axes are modified in place).
        Should use either wn_range OR wl_range, not both.
        
        To be selected, the whole bin must be inside the range.

        Parameters
        ----------
            wn_range: array, np.ndarray
                Wavenumber range in cm^-1.
            wl_range: array, np.ndarray
                Wavelength range in micron.

        Returns
        -------
            tuple:
                iw_min, iw_max the boundary indices of the spectral range
        """
        if (wn_range is None) and (wl_range is None): return None, None
        if wl_range is not None:
            if wn_range is not None:
                raise RuntimeError('Should provide either wn_range or wl_range, not both!')
            _wn_range=np.sort(10000./np.array(wl_range))
        else:
            _wn_range=np.sort(np.array(wn_range))
        iw_min=np.searchsorted(self.wnedges, _wn_range[0], side='left')
        iw_max=np.searchsorted(self.wnedges, _wn_range[1], side='right')
        iw_max-=1
        if iw_max <= iw_min:
            raise RuntimeError(f"Spectral range {wn_range} does not contain any point.")
        self.wnedges=self.wnedges[iw_min:iw_max+1]
        self.wns=self.wns[iw_min:iw_max]
        self.Nw=self.wns.size
        return iw_min, iw_max

    def blackbody(self, Temperature, integral=True):
        """Computes the surface black body flux (in W/m^2/cm^-1) at Temperature.

        Parameters
        ----------
            Temperature; float
                Blackbody temperature
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version.
        Returns
        -------
            Spectrum object
                Spectral flux in W/m^2/cm^-1
        """
        if integral:
            try:
                piB = PI*Bnu_integral_num(self.wnedges, Temperature)/np.diff(self.wnedges)
            except:
                raise RuntimeError('You should use integral=False as this spectral object does not have a wnedges atribute.')
        else:
            try:
                piB = PI*Bnu(self.wns[:], Temperature)
            except:
                raise RuntimeError('You should instenciate wns for this spectral object before calling blackbody.')
        return piB



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .util.spectral_object import Spectral_object

class Aerosols(Spectral_object):
    """Dict-like class to handle aerosol composition, reff, and density, and link this to the optical propoerties.
    """

    def __init__(self, aer_reffs_densities=None, a_database=None):
        """__init__ Instantiates an aerosols object.
        """
        super().__init__()
        self.set_a_database(a_database=a_database)
        self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)
        self._wn_range=None

    def set_aer_reffs_densities(self, aer_reffs_densities=None):
        """Set the dictionary with the arrays of the effective radii and number densities for the aerosols.

        .. warning::
            The unit and physical meaning of the result may change depending on the
            quantity provided in aer_reffs_densities (number density or ratio of particle to gas number density).
            See :func:`exo_k.atable.Atable.absorption_coefficient` for further details.

            For use in the amtospheric model, one should provide the particle to gas number density ratio.

        Parameters
        ----------
            aer_reffs_densities: dict
                A dictionary with aerosol names as keys and lists containing 2
                floats (or arrays) as values. The values are the particle effective radii
                and number densities (or ratio of aerosol to gas number density).
        """
        if aer_reffs_densities is None:
            self.aer_reffs_densities = {}
        else:
            self.aer_reffs_densities = aer_reffs_densities

    def set_a_database(self, a_database=None):
        """Change the radiative database attached to the current instance of aerosols

        Parameters
        ----------
            a_database: :class:`~exo_k.kdatabase.Kdatabase` object
                New Adatabase to use.
        """
        self.adatabase = a_database
        if self.adatabase is not None:
            if self.adatabase.r_eff_unit != 'm' :
                print("""
                You're being Bad!!! You are trying *NOT* to use MKS units
                for the effective radii in the aerosol database!!!""")
                raise RuntimeError("Bad units in the Adatabase used with aerosols.")

    def optical_properties(self, aer_reffs_densities=None, wl_range=None, wn_range=None,
            log_interp=True, compute_all_opt_prop=True, **kwargs):
        """Compute the optical properties for the mix.

        Parameters
        ----------
            aer_reffs_densities: dict
                A dictionary with aerosol names as keys and lists containing 2
                floats (or arrays) as values. The values are the particle effective radii
                and number densities (or ratio of aerosol to gas number density).
                See :func:`exo_k.atable.Atable.absorption_coefficient` for further details.

        Other Parameters
        ----------------
            wl_range, wn_range: two-value list
                Wavelength or wavenumber range to consider
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
            compute_all_opt_prop: bool, optional
                Whether to compute all the optical properties or just the extinction.
        """
        if self.adatabase is None: raise RuntimeError("""
            a_database not provided. 
            Use the a_database keyword during initialization or use the set_a_database method.""")
        if self.adatabase.wns is None: raise RuntimeError("""
            All tables in the Adatabase should have the same wavenumber grid to proceed.
            You should probably use sample().""")
        if aer_reffs_densities is not None:
            self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)

        local_wn_range=self._compute_spectral_range(wl_range=wl_range, wn_range=wn_range)

        [k, k_scat, g] = self.adatabase.optical_properties(self.aer_reffs_densities,
            wngrid_limit=local_wn_range, log_interp=log_interp,
            compute_all_opt_prop=compute_all_opt_prop)
        return [k, k_scat, g]
    
    def absorption_coefficient(self, aer_reffs_densities=None, wl_range=None, wn_range=None,
            log_interp=True, **kwargs):
        """Compute the aerosol opacity for the mix.

        Parameters
        ----------
            aer_reffs_densities: dict
                A dictionary with aerosol names as keys and lists containing 2
                floats (or arrays) as values. The values are the particle effective radii
                and number densities (or ratio of aerosol to gas number density).
                See :func:`exo_k.atable.Atable.absorption_coefficient` for further details.

        Other Parameters
        ----------------
            wl_range, wn_range: two-value list
                Wavelength or wavenumber range to consider
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).
        """
        if self.adatabase is None: raise RuntimeError("""
            a_database not provided. 
            Use the a_database keyword during initialization or use the set_a_database method.""")
        if self.adatabase.wns is None: raise RuntimeError("""
            All tables in the Adatabase should have the same wavenumber grid to proceed.
            You should probably use sample().""")
        if aer_reffs_densities is not None:
            self.set_aer_reffs_densities(aer_reffs_densities=aer_reffs_densities)

        local_wn_range=self._compute_spectral_range(wl_range=wl_range, wn_range=wn_range)

        return self.adatabase.absorption_coefficient(self.aer_reffs_densities,
            wngrid_limit=local_wn_range, log_interp=log_interp)

